import pandas as pd
import functools
from PAMI.highUtilitySpatialPattern.basic.abstract import *


class Element:
    """
    A class represents an Element of a utility list as used by the HDSHUI algorithm.

    Attributes :
    ----------
        ts : int
            keep tact of transaction id
        snu : int
            Spatial non closed itemSet utility
        remainingUtility : int
            Spatial non closed remaining utility
        pu : int
            prefix utility
        prevPos: int
            position of previous item in the list
    """

    def __init__(self, ts, snu, remainingUtility, pu, prevPos):
        self.ts = ts
        self.snu = snu
        self.remainingUtility = remainingUtility
        self.pu = pu
        self.prevPos = prevPos


class CUList:
    """
        A class represents a UtilityList as used by the HDSHUI algorithm.

    Attributes :
    ----------
        item: int
            item 
        sumSnu: long
            the sum of item utilities
        sumRemainingUtility: long
            the sum of remaining utilities
        sumCu : long
            the sum of closed utilities
        sumCru: long
            the sum of closed remaining utilities
        sumCpu: long
            the sum of closed prefix utilities
        elements: list
            the list of elements 

    Methods :
    -------
        addElement(element)
            Method to add an element to this utility list and update the sums at the same time.

    """

    def __init__(self, item):
        self.item = item
        self.sumSnu = 0
        self.sumRemainingUtility = 0
        self.sumCu = 0
        self.sumCru = 0
        self.sumCpu = 0
        self.elements = []

    def addElements(self, element):
        """
            A method to add new element to CUList
            :param element: element to be added to CUList
            :type element: Element
        """
        self.sumSnu += element.snu
        self.sumRemainingUtility += element.remainingUtility
        self.elements.append(element)


class Pair:
    """
        A class represent an item and its utility in a transaction
    """

    def __init__(self):
        self.item = 0
        self.utility = 0


class HDSHUIM(utilityPatterns):
    """
        Spatial High Utility ItemSet Mining (SHUIM) [3] is an important model in data
        mining with many real-world applications. It involves finding all spatially interesting itemSets having high value 
        in a quantitative spatio temporal database.

    Attributes :
    ----------
        iFile : str
            Name of the input file to mine complete set of frequent patterns
        oFile : str
            Name of the output file to store complete set of frequent patterns
        nFile: str
           Name of Neighbourhood items file
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program
        startTime:float
            To record the start time of the mining process
        endTime:float
            To record the completion time of the mining process
        minUtil : int
            The user given minUtil
        mapFMAP: list
            EUCS map of the FHM algorithm
        candidates: int
            candidates generated
        huiCnt: int
            huis created
        neighbors: map
            keep track of neighbours of elements
        mapOfPMU: map
            a map to keep track of Probable Maximum utilty(PMU) of each item
    Methods :
    -------
            startMine()
                Mining process will start from here
            getPatterns()
                Complete set of patterns will be retrieved with this function
            savePatterns(oFile)
                Complete set of frequent patterns will be loaded in to a output file
            constructCUL(x, compactUList, st, minUtil, length, exneighbours)
                A method to construct CUL's database
            getPatternsAsDataFrame()
                Complete set of frequent patterns will be loaded in to a dataframe
            getMemoryUSS()
                Total amount of USS memory consumed by the mining process will be retrieved from this function
            getMemoryRSS()
                Total amount of RSS memory consumed by the mining process will be retrieved from this function
            getRuntime()
                Total amount of runtime taken by the mining process will be retrieved from this function
            Explore_SearchTree(prefix, uList, exNeighbours, minUtil)
                A method to find all high utility itemSets
            updateClosed(x, compactUList, st, exCul, newT, ex, eyTs, length)
                A method to update closed values
            saveItemset(prefix, prefixLen, item, utility)
               A method to save itemSets
            updateElement(z, compactUList, st, exCul, newT, ex, duPrevPos, eyTs)
               A method to updates vales for duplicates


    Executing the code on terminal :
    -------
        Format:
            python3 HDSHUIM.py <inputFile> <outputFile> <Neighbours> <minUtil>

            python3 HDSHUIM.py <inputFile> <outputFile> <Neighbours> <minUtil> <separator>
        Examples:
            python3 HDSHUIM.py sampleTDB.txt output.txt sampleN.txt 35 (separator will be "\t" in both input and neighbourhood file)

            python3 HDSHUIM.py sampleTDB.txt output.txt sampleN.txt 35 , (separator will be "," in both input and neighbourhood file)

    Sample run of importing the code:
    -------------------------------
        
        from PAMI.highUtilityFrequentSpatialPattern.basic import HDSHUIM as alg

        obj=alg.HDSHUIM("input.txt","Neighbours.txt",35)

        obj.startMine()

        Patterns = obj.getPatterns()

        print("Total number of Spatial High-Utility Patterns:", len(Patterns))

        obj.savePatterns("output")

        memUSS = obj.getMemoryUSS()

        print("Total Memory in USS:", memUSS)

        memRSS = obj.getMemoryRSS()

        print("Total Memory in RSS", memRSS)

        run = obj.getRuntime()

        print("Total ExecutionTime in seconds:", run)

    Credits:
    -------
        The complete program was written by B.Sai Chitra under the supervision of Professor Rage Uday Kiran.
            
    """

    startTime = float()
    endTime = float()
    minSup = str()
    maxPer = float()
    finalPatterns = {}
    iFile = " "
    oFile = " "
    nFile = " "
    minUtil = 0
    memoryUSS = float()
    memoryRSS = float()
    sep = "\t"

    def __init__(self, iFile1, neighb1, minUtil, sep="\t"):
        super().__init__(iFile1, neighb1, minUtil, sep)
        self.startTime = 0
        self.endTime = 0
        self.hui_cnt = 0
        self.candidates = 0
        self.mapOfPMU = {}
        self.mapFMAP = {}
        self.neighbors = {}
        self.finalPatterns = {}

    def compareItems(self, o1, o2):
        """
            A method to sort  list of huis in pmu ascending order
        """
        compare = self.mapOfPMU[o1.item] - self.mapOfPMU[o2.item]
        if compare == 0:
            return int(o1.item) - int(o2.item)
        else:
            return compare

    def startMine(self):
        """main program to start the operation
        """
        minUtil = self.minUtil
        self.startTime = time.time()
        with open(self.nFile, 'r') as file1:
            for line in file1:
                line = line.split("\n")[0]
                parts = line.split(self.sep)
                item = parts[0]
                neigh1 = set()
                for i in range(1, len(parts)):
                    neigh1.add(parts[i])
                self.neighbors[item] = neigh1
        with open(self.iFile, 'r') as file:
            for line in file:
                parts = line.split(":")
                items_str = (parts[0].split("\n")[0]).split(self.sep)
                utility_str = (parts[2].split("\n")[0]).split(self.sep)
                transUtility = int(parts[1])
                trans1 = set()
                for i in range(0, len(items_str)):
                    trans1.add(items_str[i])
                for i in range(0, len(items_str)):
                    item = items_str[i]
                    twu = self.mapOfPMU.get(item)
                    if (twu == None):
                        twu = int(utility_str[i])
                    else:
                        twu += int(utility_str[i])
                    self.mapOfPMU[item] = twu
                    if (self.neighbors.get(item) == None):
                        continue
                    neighbours2 = trans1.intersection(self.neighbors.get(item))
                    for item2 in neighbours2:
                        if (self.mapOfPMU.get(item2) == None):
                            self.mapOfPMU[item2] = int(utility_str[i])
                        else:
                            self.mapOfPMU[item2] += int(utility_str[i])

        listOfCUList = []
        hashTable = {}
        mapItemsToCUList = {}
        for item in self.mapOfPMU.keys():
            if (self.mapOfPMU.get(item) >= minUtil):
                uList = CUList(item)
                mapItemsToCUList[item] = uList
                listOfCUList.append(uList)
        listOfCUList.sort(key=functools.cmp_to_key(self.compareItems))
        ts = 1
        with open(self.iFile, 'r') as file:
            for line in file:
                parts = line.split(":")
                items = (parts[0].split("\n")[0]).split(self.sep)
                utilities = (parts[2].split("\n")[0]).split(self.sep)
                ru = 0
                newTwu = 0
                tx_key = []
                revisedTrans = []
                for i in range(0, len(items)):
                    pair = Pair()
                    pair.item = items[i]
                    pair.utility = int(utilities[i])
                    if self.mapOfPMU.get(pair.item) >= minUtil:
                        revisedTrans.append(pair)
                        tx_key.append(pair.item)
                        newTwu += pair.utility
                revisedTrans.sort(key=functools.cmp_to_key(self.compareItems))
                tx_key1 = tuple(tx_key)
                if len(revisedTrans) > 0:
                    if tx_key1 not in hashTable.keys():
                        hashTable[tx_key1] = len(mapItemsToCUList[revisedTrans[len(revisedTrans) - 1].item].elements)
                        for i in range(len(revisedTrans) - 1, -1, -1):
                            pair = revisedTrans[i]
                            cuListOfItems = mapItemsToCUList.get(pair.item)
                            element = Element(ts, pair.utility, ru, 0, 0)
                            if i > 0:
                                element.prevPos = len(mapItemsToCUList[revisedTrans[i - 1].item].elements)
                            else:
                                element.prevPos = -1
                            cuListOfItems.addElements(element)
                            ru += pair.utility
                    else:
                        pos = hashTable[tx_key1]
                        ru = 0
                        for i in range(len(revisedTrans) - 1, -1, -1):
                            cuListOfItems = mapItemsToCUList[revisedTrans[i].item]
                            cuListOfItems.elements[pos].snu += revisedTrans[i].utility
                            cuListOfItems.elements[pos].remainingUtility += ru
                            cuListOfItems.sumSnu += revisedTrans[i].utility
                            cuListOfItems.sumRemainingUtility += ru
                            ru += revisedTrans[i].utility
                            pos = cuListOfItems.elements[pos].prevPos
                # EUCS
                for i in range(len(revisedTrans) - 1, -1, -1):
                    pair = revisedTrans[i]
                    mapFMAPItem = self.mapFMAP.get(pair.item)
                    if mapFMAPItem is None:
                        mapFMAPItem = {}
                        self.mapFMAP[pair.item] = mapFMAPItem
                    for j in range(i + 1, len(revisedTrans)):
                        pairAfter = revisedTrans[j]
                        twuSUm = mapFMAPItem.get(pairAfter.item)
                        if twuSUm is None:
                            mapFMAPItem[pairAfter.item] = newTwu
                        else:
                            mapFMAPItem[pairAfter.item] = twuSUm + newTwu
                ts += 1
        exNeighbours = set(self.mapOfPMU.keys())
        # print(self.Neighbours)
        self.Explore_SearchTree([], listOfCUList, exNeighbours, minUtil)
        self.endTime = time.time()
        process = psutil.Process(os.getpid())
        self.memoryUSS = process.memory_full_info().uss
        self.memoryRSS = process.memory_info().rss

    def Explore_SearchTree(self, prefix, uList, exNeighbours, minUtil):
        """
            A method to find all high utility itemSets

            Attributes:
            -----------
            :parm prefix: it represent all items in prefix
            :type prefix :list
            :parm uList:projected Utility list
            :type uList: list
            :parm exNeighbours: keep track of common Neighbours
            :type exNeighbours: set
            :parm minUtil:user minUtil
            :type minUtil:int
        """
        for i in range(0, len(uList)):
            x = uList[i]
            if x.item not in exNeighbours:
                continue
            self.candidates += 1
            sortedPrefix = [0] * (len(prefix) + 1)
            sortedPrefix = prefix[0:len(prefix) + 1]
            sortedPrefix.append(x.item)
            if (x.sumSnu + x.sumCu >= minUtil) and (x.item in exNeighbours):
                self.saveItemset(prefix, len(prefix), x.item, x.sumSnu + x.sumCu)
            if x.sumSnu + x.sumCu + x.sumRemainingUtility + x.sumCru >= minUtil:  # U-Prune # and (x.item in exNeighbours)):
                ULIST = []
                for j in range(i, len(uList)):
                    if (uList[j].item in exNeighbours) and (self.neighbors.get(x.item) != None) and (
                            uList[j].item in self.neighbors.get(x.item)):
                        ULIST.append(uList[j])
                exULs = self.constructCUL(x, ULIST, -1, minUtil, len(sortedPrefix), exNeighbours)
                if self.neighbors.get(x.item) != None and exNeighbours != None:
                    set1 = exNeighbours.intersection(self.neighbors.get(x.item))
                    if exULs == None or set1 == None:
                        continue
                    self.Explore_SearchTree(sortedPrefix, exULs, set1, minUtil)

    def constructCUL(self, x, compactUList, st, minUtil, length, exNeighbours):
        """
            A method to construct CUL's database

            Attributes:
            -----------
            :parm x: Compact utility list
            :type x: Node
            :parm compactUList:list of Compact utility lists
            :type compactUList:list
            :parm st: starting pos of compactUList
            :type st:int
            :parm minUtil: user minUtil
            :type minUtil:int
            :parm length: length of x
            :type length:int
            :parm exNeighbours: common Neighbours
            :type exNeighbours: list
            :return: projected database of list X
            :rtype: list or set
        """
        exCul = []
        lau = []
        cUtil = []
        eyTs = []
        for i in range(0, len(compactUList)):
            uList = CUList(compactUList[i].item)
            exCul.append(uList)
            lau.append(0)
            cUtil.append(0)
            eyTs.append(0)
        sz = len(compactUList) - (st + 1)
        exSZ = sz
        for j in range(st + 1, len(compactUList)):
            mapOfTWUF = self.mapFMAP[x.item]
            if mapOfTWUF is not None:
                twuf = mapOfTWUF.get(compactUList[j].item)
                if twuf != None and twuf < minUtil or (not (exCul[j].item in exNeighbours)):
                    exCul[j] = None
                    exSZ = sz - 1
                else:
                    uList = CUList(compactUList[j].item)
                    exCul[j] = uList
                    eyTs[j] = 0
                    lau[j] = x.sumCu + x.sumCru + x.sumSnu + x.sumRemainingUtility
                    cUtil[j] = x.sumCu + x.sumCru
        hashTable = {}
        for ex in x.elements:
            newT = []
            for j in range(st + 1, len(compactUList)):
                if exCul[j] == None:
                    continue
                eyList = compactUList[j].elements
                while eyTs[j] < len(eyList) and eyList[eyTs[j]].ts < ex.ts:
                    eyTs[j] = eyTs[j] + 1
                if eyTs[j] < len(eyList) and eyList[eyTs[j]].ts == ex.ts:
                    newT.append(j)
                else:
                    lau[j] = lau[j] - ex.snu - ex.remainingUtility
                    if lau[j] < minUtil:
                        exCul[j] = None
                        exSZ = exSZ - 1
            if len(newT) == exSZ:
                self.updateClosed(x, compactUList, st, exCul, newT, ex, eyTs, length)
            else:
                if len(newT) == 0:
                    continue
                ru = 0
                newT1 = tuple(newT)
                if newT1 not in hashTable.keys():
                    hashTable[newT1] = len(exCul[newT[len(newT) - 1]].elements)
                    for i in range(len(newT) - 1, -1, -1):
                        cuListOfItems = exCul[newT[i]]
                        y = compactUList[newT[i]].elements[eyTs[newT[i]]]
                        element = Element(ex.ts, ex.snu + y.snu - ex.pu, ru, ex.snu, 0)
                        if i > 0:
                            element.prevPos = len(exCul[newT[i - 1]].elements)
                        else:
                            element.prevPos = -1
                        cuListOfItems.addElements(element)
                        ru += y.snu - ex.pu
                else:
                    dPrevPos = hashTable[newT1]
                    self.updateElement(x, compactUList, st, exCul, newT, ex, dPrevPos, eyTs)
            for j in range(st + 1, len(compactUList)):
                cUtil[j] = cUtil[j] + ex.snu + ex.remainingUtility
        filter_compactUList = []
        for j in range(st + 1, len(compactUList)):
            if cUtil[j] < minUtil or exCul[j] is None:
                continue
            else:
                if length > 1:
                    exCul[j].sumCu += compactUList[j].sumCu + x.sumCu - x.sumCpu
                    exCul[j].sumCru += compactUList[j].sumCru
                    exCul[j].sumCpu += x.sumCu
                filter_compactUList.append(exCul[j])
        return filter_compactUList

    def updateClosed(self, x, compactUList, st, exCul, newT, ex, eyTs, length):
        """
            A method to update closed values
            Attributes:
            -----------
            :parm x: Compact utility list
            :type x: list
            :parm compactUList:list of Compact utility lists
            :type compactUList:list
            :parm st: starting pos of compactUList
            :type st:int
            :parm newT:transaction to be updated
            :type newT:list
            :parm ex: element ex
            :type ex:element
            :parm eyTs:list of tss
            :type eyTs:ts
            :parm length: length of x
            :type length:int

        """
        remainingUtility = 0
        for j in range(len(newT) - 1, -1, -1):
            ey = compactUList[newT[j]]
            eyy = ey.elements[eyTs[newT[j]]]
            exCul[newT[j]].sumCu += ex.snu + eyy.snu - ex.pu
            exCul[newT[j]].sumCru += remainingUtility
            exCul[newT[j]].sumCpu += ex.snu
            remainingUtility = remainingUtility + eyy.snu - ex.pu

    def updateElement(self, z, compactUList, st, exCul, newT, ex, duPrevPos, eyTs):
        """
            A method to updates vales for duplicates

            Attributes:
            -----------
            :parm z: Compact utility list
            :type z: list
            :parm compactUList:list of Compact utility lists
            :type compactUList:list
            :parm st: starting pos of compactUList
            :type st:int
            :parm exCul:list of compactUList
            :type exCul:list
            :parm newT:transaction to be updated
            :type newT:list
            :parm ex: element ex
            :type ex:element
            :parm duPrevPos: position of z in exCul
            :type duPrevPos:int
            :parm eyTs:list of tss
            :type eyTs:ts
        """
        remainingUtility = 0
        pos = duPrevPos
        for j in range(len(newT) - 1, -1, -1):
            ey = compactUList[newT[j]]
            eyy = ey.elements[eyTs[newT[j]]]
            exCul[newT[j]].elements[pos].snu += ex.snu + eyy.snu - ex.pu
            exCul[newT[j]].sumSnu += ex.snu + eyy.snu - ex.pu
            exCul[newT[j]].elements[pos].remainingUtility += remainingUtility
            exCul[newT[j]].sumRemainingUtility += remainingUtility
            exCul[newT[j]].elements[pos].pu += ex.snu
            remainingUtility = remainingUtility + eyy.snu - ex.pu
            pos = exCul[newT[j]].elements[pos].prevPos

    def saveItemset(self, prefix, prefixLen, item, utility):
        """
         A method to save itemSets

         Attributes:
        -----------
        :parm prefix: it represent all items in prefix
        :type prefix :list
        :parm item:item
        :type item: int
        :parm utility:utility of itemset
        :type utility:int
        """
        self.hui_cnt += 1
        res = ""
        for i in range(0, prefixLen):
            res += str(prefix[i]) + " "
        res += str(item)
        res1 = str(utility)
        self.finalPatterns[res] = res1

    def getPatternsAsDataFrame(self):
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self.finalPatterns.items():
            data.append([a, b])
            dataFrame = pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataFrame

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self.finalPatterns

    def savePatterns(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file

        :param outFile: name of the output file
        :type outFile: file
        """
        self.oFile = outFile
        writer = open(self.oFile, 'w+')
        for x, y in self.finalPatterns.items():
            patternsAndSupport = str(x) + " : " + str(y)
            writer.write("%s \n" % patternsAndSupport)

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self.memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
       """
        return self.memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process


        :return: returning total amount of runtime taken by the mining process
        :rtype: float
       """
        return self.endTime - self.startTime


if __name__ == "__main__":
    ap = str()
    if len(sys.argv) == 5 or len(sys.argv) == 6:
        if len(sys.argv) == 6:  # to  include a user specified separator
            ap = HDSHUIM(sys.argv[1], sys.argv[3], int(sys.argv[4]), sys.argv[5])
        if len(sys.argv) == 5:  # to consider "\t" as a separator
            ap = HDSHUIM(sys.argv[1], sys.argv[3], int(sys.argv[4]))
        ap.startMine()
        Patterns = ap.getPatterns()
        print("Total number of Spatial High-Utility Patterns:", len(Patterns))
        ap.savePatterns(sys.argv[2])
        memUSS = ap.getMemoryUSS()
        print("Total Memory in USS:", memUSS)
        memRSS = ap.getMemoryRSS()
        print("Total Memory in RSS", memRSS)
        run = ap.getRuntime()
        print("Total ExecutionTime in ms:", run)
    else:
        ap = HDSHUIM('/home/apiiit-rkv/Downloads/ffsi_rainFallHighUtilityTransactionalDatabase.txt',
                     '/home/apiiit-rkv/Downloads/ffsi_neighborhoodRainFall_6.txt', 3000000, ' ')
        ap.startMine()
        Patterns = ap.getPatterns()
        print("Total number of Spatial High-Utility Patterns:", len(Patterns))
        ap.savePatterns(sys.argv[2])
        memUSS = ap.getMemoryUSS()
        print("Total Memory in USS:", memUSS)
        memRSS = ap.getMemoryRSS()
        print("Total Memory in RSS", memRSS)
        run = ap.getRuntime()
        print("Total ExecutionTime in ms:", run)
        print("Error! The number of input parameters do not match the total number of parameters provided")



import matplotlib.pyplot as plt

class plotLineGraphFromDictionary:
    """
    This class plot graph of input data

    Attributes:
    ----------
    data : dict
        store input data as dict

    Methods:
    -------
    plotLineGraph()
        draw line graph of input data. input data's key is x and value is y.
    """
    def __init__(self, data, percentage=100, title='', xlabel='', ylabel=''):

        """
        draw line graph. Plot the input data key as x and value as y
        :param percentage: percentage of graph to plot
        :type percentage: int
        :param title: title of graph
        :type title: str
        :param xlabel: xlabel of graph
        :type xlabel: str
        :param ylabel: ylabel of grapth
        :type ylabel: str
        """
        numberOfGraphToPlot = int(len(data) * percentage / 100)
        x = tuple(data.keys())[:numberOfGraphToPlot]
        y = tuple(data.values())[:numberOfGraphToPlot]
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.set_title(title)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.plot(x, y)
        #plt.plot(x, y)
        plt.show()


#  Copyright (C)  2021 Rage Uday Kiran
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
#  Copyright (C)  2021 Rage Uday Kiran

import sys
from urllib.request import urlopen
import validators
from PAMI.frequentPattern.basic.abstract import *


class ECLATbitset(frequentPatterns):
    """
    ECLATbitset is one of the fundamental algorithm to discover frequent patterns in a transactional database.
    This program implemented following the eclat bitset algorithm.

    Reference:
    ----------
        Zaki, M.J., Gouda, K.: Fast vertical mining using diffsets. Technical Report 01-1, Computer Science
            Dept., Rensselaer Polytechnic Institute (March 2001), https://doi.org/10.1145/956750.956788

    Attributes:
    -----------
        self.iFile : str
            Input file name or path of the input file
        minSup: float or int or str
            The user can specify minSup either in count or proportion of database size.
            If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.
            Otherwise, it will be treated as float.
            Example: minSup=10 will be treated as integer, while minSup=10.0 will be treated as float
        sep : str
            This variable is used to distinguish items from one another in a transaction. The default separator is tab space or \t.
            However, the users can override their default separator.
        self.oFile : str
            Name of the output file or path of the output file
        self.startTime:float
            To record the start time of the mining process
        self.endTime:float
            To record the completion time of the mining process
        self.finalPatterns: dict
            Storing the complete set of patterns in a dictionary variable
        self.memoryUSS : float
            To store the total amount of USS memory consumed by the program
        self.memoryRSS : float
            To store the total amount of RSS memory consumed by the program
        self.Database : list
            To store the complete set of transactions available in the input database/file

    Methods:
    -------
    startMine()
        Mining process will start from here
    getPatterns()
        Complete set of patterns will be retrieved with this function
    savePatterns(oFile)
        Complete set of frequent patterns will be loaded in to a output file
    getPatternsAsDataFrame()
        Complete set of frequent patterns will be loaded in to a dataframe
    getMemoryUSS()
        Total amount of USS memory consumed by the mining process will be retrieved from this function
    getMemoryRSS()
        Total amount of RSS memory consumed by the mining process will be retrieved from this function
    getRuntime()
        Total amount of runtime taken by the mining process will be retrieved from this function
    creatingItemSets(iFileName)
        Storing the complete transactions of the database/input file in a database variable
    generationOfAllItems()
        It will generate the combinations of frequent items
    startMine()
        the main function to mine the patterns

    Executing the code on terminal:
    -------------------------------

        Format:
        -------
        python3 ECLATbitset.py <inputFile> <outputFile> <minSup>

        Examples:
        ---------
        python3 ECLATbitset.py sampleDB.txt patterns.txt 10.0   (minSup will be considered in percentage of database transactions)

        python3 ECLATbitset.py sampleDB.txt patterns.txt 10     (minSup will be considered in support count or frequency)

    Sample run of the importing code:
    ---------------------------------

        import PAMI.frequentPattern.basic.ECLATbitset as alg

        obj = alg.ECLATbitset(iFile, minSup)

        obj.startMine()

        frequentPatterns = obj.getPatterns()

        print("Total number of Frequent Patterns:", len(frequentPatterns))

        obj.savePatterns(oFile)

        Df = obj.getPatternInDataFrame()

        memUSS = obj.getMemoryUSS()

        print("Total Memory in USS:", memUSS)

        memRSS = obj.getMemoryRSS()

        print("Total Memory in RSS", memRSS)

        run = obj.getRuntime()

        print("Total ExecutionTime in seconds:", run)

    Credits:
    --------
        The complete program was written by P.Likhitha  under the supervision of Professor Rage Uday Kiran.

        """
    startTime = float()
    endTime = float()
    finalPatterns = {}
    iFile = " "
    oFile = " "
    sep = " "
    minSup = str()
    memoryUSS = float()
    memoryRSS = float()
    Database = []
    mapSupport = {}
    lno = 0

    def convert(self, value):
        """
        To convert the user specified minSup value

        :param value: user specified minSup value

        :return: converted type
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (self.lno * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (self.lno * value)
            else:
                value = int(value)
        return value

    def creatingItemSets(self):
        """
            Storing the complete transactions of the database/input file in a database variable

        """
        self.Database = []
        self.mapSupport = {}
        if isinstance(self.iFile, pd.DataFrame):
            if self.iFile.empty:
                print("its empty..")
            i = self.iFile.columns.values.tolist()
            if 'Transactions' in i:
                self.Database = self.iFile['Transactions'].tolist()
            if 'Patterns' in i:
                self.Database = self.iFile['Patterns'].tolist()
        if isinstance(self.iFile, str):
            if validators.url(self.iFile):
                data = urlopen(self.iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self.sep)]
                    temp = [x for x in temp if x]
                    self.Database.append(temp)
            else:
                try:
                    with open(self.iFile, 'r') as f:
                        for line in f:
                            self.lno += 1
                            splitter = [i.rstrip() for i in line.split(self.sep)]
                            splitter = [x for x in splitter if x]
                            self.Database.append(splitter)
                except IOError:
                    print("File Not Found")
        self.minSup = self.convert(self.minSup)
    
    def OneFrequentItems(self):
        items = []
        p = {}
        for i in self.Database:
            for j in i:
                if j not in items:
                    items.append(j)
        for temp in self.Database:
            for j in items:
                count = 0
                if j in temp:
                    count = 1
                if j not in p:
                    p[j] = [count]
                else:
                    p[j].append(count)
        for x, y in p.items():
            if self.countSupport(y) >= self.minSup:
                self.mapSupport[x] = y
        pList = [key for key, value in sorted(self.mapSupport.items(), key=lambda x: (len(x[1])), reverse=True)]
        return pList

    @staticmethod
    def countSupport(tids):
        """To count support of 1's in tids

        :param tids: bitset representation of itemSets

        :return:  count
        """
        count = 0
        for i in tids:
            if i == 1:
                count += 1
        return count

    def save(self, prefix, suffix, tidSetX):
        """To save the patterns satisfying the minSup condition

        :param prefix: prefix item of itemSet

        :param suffix: suffix item of itemSet

        :param tidSetX: bitset representation of itemSet

        :return: saving the itemSet in to finalPatterns
        """
        if prefix is None:
            prefix = suffix
        else:
            prefix = prefix + suffix
        count = self.countSupport(tidSetX)
        sample = str()
        for i in prefix:
            sample = sample + i + " "
        self.finalPatterns[sample] = count

    def generationOfAll(self, prefix, itemSets, tidSets):
        """It will generate the combinations of frequent items with prefix and  list of items

            :param prefix: it represents the prefix item to form the combinations

            :type prefix: list

            :param itemSets: it represents the suffix items of prefix

            :type itemSets: list

            :param tidSets: represents the tidLists of itemSets

            :type tidSets: 2d list
        """
        if len(itemSets) == 1:
            i = itemSets[0]
            tidI = tidSets[0]
            self.save(prefix, [i], tidI)
            return
        for i in range(len(itemSets)):
            itemI = itemSets[i]
            if itemI is None:
                continue
            tidSetX = tidSets[i]
            classItemSets = []
            classTidSets = []
            itemSetx = [itemI]
            for j in range(i + 1, len(itemSets)):
                itemJ = itemSets[j]
                tidSetJ = tidSets[j]
                y = [k & l for k, l in zip(tidSetX, tidSetJ)]
                support = self.countSupport(y)
                if support >= self.minSup:
                    classItemSets.append(itemJ)
                    classTidSets.append(y)
            newprefix = list(set(itemSetx)) + prefix
            self.generationOfAll(newprefix, classItemSets, classTidSets)
            del classItemSets, classTidSets
            self.save(prefix, list(set(itemSetx)), tidSetX)
            #raise Exception("end of time")

    def startMine(self):
        """Frequent pattern mining process will start from here
        We start with the scanning the itemSets and store the bitsets respectively.
        We form the combinations of single items and  check with minSup condition to check the frequency of patterns
        """

        self.startTime = time.time()
        if self.iFile is None:
            raise Exception("Please enter the file path or file name:")
        if self.minSup is None:
            raise Exception("Please enter the Minimum Support")
        self.creatingItemSets()
        plist = self.OneFrequentItems()
        self.finalPatterns = {}
        for i in range(len(plist)):
            itemI = plist[i]
            tidSetX = self.mapSupport[itemI]
            itemSetx = [itemI]
            itemSets = []
            tidSets = []
            for j in range(i + 1, len(plist)):
                itemJ = plist[j]
                tidSetJ = self.mapSupport[itemJ]
                y1 = [k & l for k, l in zip(tidSetX, tidSetJ)]
                support = self.countSupport(y1)
                if support >= self.minSup:
                    itemSets.append(itemJ)
                    tidSets.append(y1)
            self.generationOfAll(itemSetx, itemSets, tidSets)
            del itemSets, tidSets
            self.save(None, itemSetx, tidSetX)
        self.endTime = time.time()
        process = psutil.Process(os.getpid())
        self.memoryUSS = float()
        self.memoryRSS = float()
        self.memoryUSS = process.memory_full_info().uss
        self.memoryRSS = process.memory_info().rss
        print("Frequent patterns were generated successfully using Eclat_bitset algorithm")

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process

        :rtype: float
        """

        return self.memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process

        :rtype: float
        """

        return self.memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process

        :rtype: float
        """

        return self.endTime - self.startTime

    def getPatternsAsDataFrame(self):
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe

        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self.finalPatterns.items():
            data.append([a, b])
            dataFrame = pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataFrame

    def savePatterns(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file

        :param outFile: name of the output file

        :type outFile: file
        """
        self.oFile = outFile
        writer = open(self.oFile, 'w+')
        for x, y in self.finalPatterns.items():
            patternsAndSupport = x + ":" + str(y)
            writer.write("%s \n" % patternsAndSupport)

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns

        :rtype: dict
        """
        return self.finalPatterns


if __name__ == "__main__":
    ap = str()
    if len(sys.argv) == 4 or len(sys.argv) == 5:
        if len(sys.argv) == 5:
            ap = ECLATbitset(sys.argv[1], sys.argv[3], sys.argv[4])
        if len(sys.argv) == 4:
            ap = ECLATbitset(sys.argv[1], sys.argv[3])
        ap.startMine()
        Patterns = ap.getPatterns()
        print("Total number of Frequent Patterns:", len(Patterns))
        ap.savePatterns(sys.argv[2])
        memUSS = ap.getMemoryUSS()
        print("Total Memory in USS:", memUSS)
        memRSS = ap.getMemoryRSS()
        print("Total Memory in RSS", memRSS)
        run = ap.getRuntime()
        print("Total ExecutionTime in ms:", run)
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")



import sys
from PAMI.highUtilityPatterns.basic.abstract import *


class Element:
    """
    A class represents an Element of a utility list .

    Attributes :
    ----------
        ts : int
            keep tact of transaction id
        nu : int
            non closed itemSet utility
        nru : int
             non closed remaining utility
        pu : int
            prefix utility
        ppos: int
            position of previous item in the list
    """

    def __init__(self, tid, nu, nru, pu, ppos):
        self.tid = tid
        self.nu = nu
        self.nru = nru
        self.pu = pu
        self.ppos = ppos


class CUList:
    """
        A class represents a UtilityList

    Attributes :
    ----------
        item: int
            item 
        sumNu: long
            the sum of item utilities
        sumNru: long
            the sum of remaining utilities
        sumCu : long
            the sum of closed utilities
        sumCru: long
            the sum of closed remaining utilities
        sumCpu: long
            the sum of closed prefix utilities
        elements: list
            the list of elements 

    Methods :
    -------
        addElement(element)
            Method to add an element to this utility list and update the sums at the same time.
    """
    def __init__(self, item):
        self.item = item
        self.sumnu = 0
        self.sumnru = 0
        self.sumCu = 0
        self.sumCru = 0
        self.sumCpu = 0
        self.elements = []

    def addElements(self, element):
        """
            A method to add new element to CUList
            :param element: element to be addeed to CUList
            :type element: Element
        """
        self.sumnu += element.nu
        self.sumnru += element.nru
        self.elements.append(element)


class Pair:
    """
        A class represent an item and its utility in a transaction
    """
    def __init__(self):
        self.item = 0
        self.utility = 0


class HMiner(utilityPatterns):
    """
        High Utility itemSet Mining (HMIER) is an importent algorithm to miner High utility items from the database.

    Attributes:
    ----------
        iFile : file
            Name of the input file to mine complete set of frequent patterns
        oFile : file
            Name of the output file to store complete set of frequent patterns
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program
        startTime:float
            To record the start time of the mining process
        endTime:float
            To record the completion time of the mining process
        minUtil : int
            The user given minUtil
        mapFMAP: list
            EUCS map of the FHM algorithm
        candidates: int
            candidates genetated
        huiCnt: int
            huis created
        neighbors: map
            keep track of nighboues of elements
    Methods:
    -------
        startMine()
            Mining process will start from here
        getPatterns()
            Complete set of patterns will be retrieved with this function
        savePatterns(oFile)
            Complete set of frequent patterns will be loaded in to a output file
        getPatternsAsDataFrame()
            Complete set of frequent patterns will be loaded in to a dataframe
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function
        Explore_SearchTree(prefix, uList, minUtil)
            A method to find all high utility itemSets
        UpdateCLosed(x, culs, st, excul, newT, ex, ey_ts, length)
            A method to update closed values
        saveitemSet(prefix, prefixLen, item, utility)
            A method to save itemSets
        updateElement(z, culs, st, excul, newT, ex, duppos, ey_ts)
            A method to updates vales for duplicates
        construcCUL(x, culs, st, minUtil, length, exnighbors)
            A method to construct CUL's database

    Executing the code on terminal :
    -------
        Format:
            python3 HMiner.py <inputFile> <outputFile> <minUtil>

            python3 HMiner.py <inputFile> <outputFile> <minUtil> <separator>
        Examples:
            python3 HMiner.py sampleTDB.txt output.txt 35 (separator will be "\t")

            python3 HMiner.py sampleTDB.txt output.txt 35 ,  (separator will be "," in input file)

    Sample run of importing the code:
    -------------------------------
        
        from PAMI.highUtilityPatterns.basic import HMiner as alg

        obj = alg.HMiner("input.txt",35)

        obj.startMine()

        Patterns = obj.getPatterns()

        print("Total number of high utility Patterns:", len(Patterns))

        obj.savePatterns("output")

        memUSS = obj.getMemoryUSS()

        print("Total Memory in USS:", memUSS)

        memRSS = obj.getMemoryRSS()

        print("Total Memory in RSS", memRSS)

        run = obj.getRuntime()

        print("Total ExecutionTime in seconds:", run)

    Credits:
    -------
        The complete program was written by B.Sai Chitra under the supervision of Professor Rage Uday Kiran.
            
    """
    
    startTime = float()
    endTime = float()
    minSup = str()
    maxPer = float()
    finalPatterns = {}
    Database = {}
    transactions = []
    utilities = []
    utilitySum = []
    iFile = " "
    oFile = " "
    minUtil = 0
    sep = "\t"
    memoryUSS = float()
    memoryRSS = float()

    def __init__(self, iFile1, minUtil, sep="\t"):
        super().__init__(iFile1, minUtil, sep)
        self.start = 0
        self.end = 0
        self.hui_cnt = 0
        self.candidates = 0
        self.mapOfTWU = {}
        self.minutil = 0
        self.mapFMAP = {}
        self.finalPatterns = {}

    def HMiner(self, o1, o2):
        """
            A method to sort  list of huis in TWU asending order
        """
        compare = self.mapOfTWU[o1.item] - self.mapOfTWU[o2.item]
        if compare == 0:
            return int(o1.item) - int(o2.item)
        else:
            return compare

    def creteItemsets(self):
        self.transactions, self.utilities, self.utilitySum = [], [], []
        if isinstance(self.iFile, pd.DataFrame):
            if self.iFile.empty:
                print("its empty..")
            i = self.iFile.columns.values.tolist()
            if 'Transactions' in i:
                self.transactions = self.iFile['Transactions'].tolist()
            if 'Utilities' in i:
                self.utilities = self.iFile['Utilities'].tolist()
            if 'UtilitySum' in i:
                self.utilitySum = self.iFile['UtilitySum'].tolist()
        if isinstance(self.iFile, str):
            if validators.url(self.iFile):
                print("hey")
                data = urlopen(self.iFile)
                for line in data:
                    line = line.decode("utf-8")
                    line = line.split("\n")[0]
                    parts = line.split(":")
                    items = parts[0].split(self.sep)
                    self.transactions.append([x for x in items if x])
                    utilities = parts[2].split(self.sep)
                    self.utilities.append(utilities)
                    self.utilitySum.append(int(parts[1]))
            else:
                try:
                    with open(self.iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line = line.split("\n")[0]
                            parts = line.split(":")
                            items = parts[0].split(self.sep)
                            self.transactions.append([x for x in items if x])
                            utilities = parts[2].split(self.sep)
                            self.utilities.append(utilities)
                            self.utilitySum.append(int(parts[1]))
                except IOError:
                    print("File Not Found")
                    quit()


    def startMine(self):
        """
            main program to start the operation
        """
        self.startTime = time.time()
        self.creteItemsets()
        self.finalPatterns = {}
        for line in range(len(self.transactions)):
            items_str = self.transactions[line]
            utility_str = self.utilities[line]
            transUtility = self.utilitySum[line]
            for i in range(0, len(items_str)):
                item = items_str[i]
                twu = self.mapOfTWU.get(item)
                if twu == None:
                    twu = transUtility
                else:
                    twu += transUtility
                self.mapOfTWU[item] = twu
        listOfCUList = []
        hashTable = {}
        mapItemsToCUList = {}
        minutil = self.minUtil
        for item in self.mapOfTWU.keys():
            if self.mapOfTWU.get(item) >= minutil:
                uList = CUList(item)
                mapItemsToCUList[item] = uList
                listOfCUList.append(uList)
        listOfCUList.sort(key=functools.cmp_to_key(self.HMiner))
        tid = 1
        for line in range(len(self.transactions)):
            items = self.transactions[line]
            utilities = self.utilities[line]
            ru = 0
            newTwu = 0
            tx_key = []
            revisedTrans = []
            for i in range(0, len(items)):
                pair = Pair()
                pair.item = items[i]
                pair.utility = int(utilities[i])
                if self.mapOfTWU.get(pair.item) >= minutil:
                    revisedTrans.append(pair)
                    tx_key.append(pair.item)
                    newTwu += pair.utility
            revisedTrans.sort(key=functools.cmp_to_key(self.HMiner))
            tx_key1 = tuple(tx_key)
            if len(revisedTrans) > 0:
                if tx_key1 not in hashTable.keys():
                    hashTable[tx_key1] = len(mapItemsToCUList[revisedTrans[len(revisedTrans)-1].item].elements)
                    for i in range(len(revisedTrans)-1, -1, -1):
                        pair = revisedTrans[i]
                        cuListoFItems = mapItemsToCUList.get(pair.item)
                        element = Element(tid, pair.utility, ru, 0, 0)
                        if i > 0:
                            element.ppos = len(mapItemsToCUList[revisedTrans[i-1].item].elements)
                        else:
                            element.ppos =- 1
                        cuListoFItems.addElements(element)
                        ru += pair.utility
                else:
                    pos = hashTable[tx_key1]
                    ru = 0
                    for i in range(len(revisedTrans)-1, -1, -1):
                        cuListoFItems = mapItemsToCUList[revisedTrans[i].item]
                        cuListoFItems.elements[pos].nu += revisedTrans[i].utility
                        cuListoFItems.elements[pos].nru += ru
                        cuListoFItems.sumnu += revisedTrans[i].utility
                        cuListoFItems.sumnru += ru
                        ru += revisedTrans[i].utility
                        pos = cuListoFItems.elements[pos].ppos
                    #EUCS
            for i in range(len(revisedTrans)-1, -1, -1):
                pair = revisedTrans[i]
                mapFMAPItem = self.mapFMAP.get(pair.item)
                if mapFMAPItem == None:
                    mapFMAPItem = {}
                    self.mapFMAP[pair.item] = mapFMAPItem
                for j in range(i+1, len(revisedTrans)):
                    pairAfter = revisedTrans[j]
                    twuSUm = mapFMAPItem.get(pairAfter.item)
                    if twuSUm is None:
                        mapFMAPItem[pairAfter.item] = newTwu
                    else:
                        mapFMAPItem[pairAfter.item] = twuSUm + newTwu
            tid += 1
        self.Explore_SearchTree([], listOfCUList, minutil)
        self.endTime = time.time()
        process = psutil.Process(os.getpid())
        self.memoryUSS = process.memory_full_info().uss
        self.memoryRSS = process.memory_info().rss
        print("High Utility patterns were generated successfully using HMiner algorithm")

    def Explore_SearchTree(self, prefix, uList, minutil):
        """
            A method to find all high utility itemSets

            Attributes:
            -----------
            :parm prefix: it represent all items in prefix
            :type prefix :list
            :parm uList:projectd Utility list
            :type uList: list
            :parm minutil:user minUtil
            :type minutil:int
        """
        for i in range(0, len(uList)):
            x = uList[i]
            soted_prefix = [0] * (len(prefix) + 1)
            soted_prefix = prefix[0:len(prefix) + 1]
            soted_prefix.append(x.item)
            if x.sumnu + x.sumCu >= minutil:
                self.saveitemSet(prefix, len(prefix), x.item, x.sumnu + x.sumCu)
            self.candidates += 1
            if x.sumnu + x.sumCu + x.sumnru + x.sumCru >= minutil:
                exULs = self.construcCUL(x, uList, i, minutil, len(soted_prefix))
                self.Explore_SearchTree(soted_prefix, exULs, minutil)

    def construcCUL(self, x, culs, st, minutil, length):
        """
            A method to construct CUL's database

            Attributes:
            -----------
            :parm x: Compact utility list
            :type x: Node
            :parm culs:list of Compact utility lists
            :type culs:list
            :parm st: starting pos of culs
            :type st:int
            :parm minutil: user minUtil
            :type minutil:int
            :parm length: length of x
            :type length:int
            :return: projectd database of list X
            :rtype: list
        """
        excul = []
        lau = []
        cutil = []
        ey_tid = []
        for i in range(0, len(culs)):
            uList = CUList(culs[i].item)
            excul.append(uList)
            lau.append(0)
            cutil.append(0)
            ey_tid.append(0)
        sz = len(culs)-(st+1)
        exSZ = sz
        for j in range(st+1, len(culs)):
            mapOfTWUF = self.mapFMAP[x.item]
            if mapOfTWUF != None:
                twuf = mapOfTWUF.get(culs[j].item)
                if twuf != None and twuf < minutil:
                    excul[j] = None
                    exSZ = sz-1
                else:
                    uList = CUList(culs[j].item)
                    excul[j] = uList
                    ey_tid[j] = 0
                    lau[j] = x.sumCu + x.sumCru + x.sumnu + x.sumnru
                    cutil[j] = x.sumCu + x.sumCru
        hashTable = {}            
        for ex in x.elements:
            newT = []
            for j in range(st+1, len(culs)):
                if excul[j] is None:
                    continue
                eylist = culs[j].elements
                while ey_tid[j] < len(eylist) and eylist[ey_tid[j]].tid < ex.tid:
                    ey_tid[j] = ey_tid[j] + 1
                if ey_tid[j] < len(eylist) and eylist[ey_tid[j]].tid == ex.tid:
                    newT.append(j)
                else:
                    lau[j] = lau[j] - ex.nu - ex.nru
                    if lau[j] < minutil:
                        excul[j] = None
                        exSZ = exSZ - 1
            if len(newT) == exSZ:
                self.UpdateCLosed(x, culs, st, excul, newT, ex, ey_tid, length)
            else:
                if len(newT) == 0:
                    continue
                ru = 0
                newT1 = tuple(newT)
                if newT1 not in hashTable.keys():
                    hashTable[newT1] = len(excul[newT[len(newT)-1]].elements)
                    for i in range(len(newT)-1, -1, -1):
                        cuListoFItems = excul[newT[i]]
                        y = culs[newT[i]].elements[ey_tid[newT[i]]]
                        element = Element(ex.tid, ex.nu + y.nu - ex.pu, ru, ex.nu, 0)
                        if i > 0:
                            element.ppos = len(excul[newT[i-1]].elements)
                        else:
                            element.ppos =- 1
                        cuListoFItems.addElements(element)
                        ru += y.nu - ex.pu
                else:
                    dppos = hashTable[newT1]
                    self.updateElement(x, culs, st, excul, newT, ex, dppos, ey_tid)
            for j in range(st+1, len(culs)):
                cutil[j] = cutil[j] + ex.nu + ex.nru
        filter_culs = []
        for j in range(st+1, len(culs)):
            if cutil[j] < minutil or excul[j] is None:
                continue
            else:
                if length > 1:
                    excul[j].sumCu += culs[j].sumCu + x.sumCu - x.sumCpu
                    excul[j].sumCru += culs[j].sumCru
                    excul[j].sumCpu += x.sumCu
                filter_culs.append(excul[j])
        return filter_culs
    
    def UpdateCLosed(self, x, culs, st, excul, newT, ex, ey_tid, length):
        """
            A method to update closed values
            Attributes:
            -----------
            :parm x: Compact utility list
            :type x: list
            :parm culs:list of Compact utility lists
            :type culs:list
            :parm st: starting pos of culs
            :type st:int
            :parm newT:transaction to be updated
            :type newT:list
            :parm ex: element ex
            :type ex:element
            :parm ey_tid:list of tss
            :type ey_tid:ts
            :parm length: length of x
            :type length:int

        """
        nru = 0
        for j in range(len(newT)-1, -1, -1):
            ey = culs[newT[j]]
            eyy = ey.elements[ey_tid[newT[j]]]
            excul[newT[j]].sumCu += ex.nu + eyy.nu - ex.pu
            excul[newT[j]].sumCru += nru
            excul[newT[j]].sumCpu += ex.nu
            nru = nru + eyy.nu - ex.pu
    
    def updateElement(self, z, culs, st, excul, newT, ex, duppos, ey_tid):
        """
            A method to updates vales for duplicates

            Attributes:
            -----------
            :parm z: Compact utility list
            :type z: list
            :parm culs:list of Compact utility lists
            :type culs:list
            :parm st: starting pos of culs
            :type st:int
            :parm excul:list of culs
            :type excul:list
            :parm newT:transaction to be updated
            :type newT:list
            :parm ex: element ex
            :type ex:element
            :parm duppos: position of z in excul
            :type duppos:int
            :parm ey_tid:list of tss
            :type ey_tid:ts
        """
        nru = 0
        pos = duppos
        for j in range(len(newT)-1, -1, -1):
            ey = culs[newT[j]]
            eyy = ey.elements[ey_tid[newT[j]]]
            excul[newT[j]].elements[pos].nu += ex.nu + eyy.nu - ex.pu
            excul[newT[j]].sumnu += ex.nu + eyy.nu - ex.pu
            excul[newT[j]].elements[pos].nru += nru
            excul[newT[j]].sumnru += nru
            excul[newT[j]].elements[pos].pu += ex.nu
            nru = nru + eyy.nu - ex.pu
            pos = excul[newT[j]].elements[pos].ppos
    
    def saveitemSet(self, prefix, prefixLen, item, utility):
        """
         A method to save itemSets

         Attributes:
        -----------
        :parm prefix: it represent all items in prefix
        :type prefix :list
        :pram prefixLen: length of prefix
        :type prefixLen:int
        :parm item:item
        :type item: int
        :parm utility:utility of itemSet
        :type utility:int
        """
        self.hui_cnt += 1
        res = ""
        for i in range(0, prefixLen):
            res += str(prefix[i])+" "
        res += str(item)
        self.finalPatterns[str(res)] = str(utility)
    
    def getPatternsAsDataFrame(self):
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self.finalPatterns.items():
            data.append([a, b])
            dataFrame = pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataFrame
    
    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self.finalPatterns
    
    def savePatterns(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file

        :param outFile: name of the output file
        :type outFile: file
        """
        self.oFile = outFile
        writer = open(self.oFile, 'w+')
        for x, y in self.finalPatterns.items():
            patternsAndSupport = str(x) + " : " + str(y)
            writer.write("%s\n" % patternsAndSupport)
    
    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self.memoryUSS
    
    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
       """
        return self.memoryRSS
    
    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process


        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """
        return self.endTime-self.startTime


if __name__ == "__main__":
    ap = str()
    if len(sys.argv) == 4 or len(sys.argv) == 5:
        if len(sys.argv) == 5: #includes separator
            ap = HMiner(sys.argv[1], int(sys.argv[3]), sys.argv[4])
        if len(sys.argv) == 4: # to consider "\t" as aseparator
            ap = HMiner(sys.argv[1], int(sys.argv[3]))
        ap.startMine()
        Patterns = ap.getPatterns()
        print("Total number of huis:", len(Patterns))
        ap.savePatterns(sys.argv[2])
        memUSS = ap.getMemoryUSS()
        print("Total Memory in USS:", memUSS)
        memRSS = ap.getMemoryRSS()
        print("Total Memory in RSS", memRSS)
        run = ap.getRuntime()
        print("Total ExecutionTime in ms:", run)
    else:
        l = [200000, 300000, 400000, 500000]
        for i in l:
            ap = HMiner('/home/apiiit-rkv/Downloads/Reaserch/maximal/mushroom_utility_SPMF.txt', i, ' ')
            ap.startMine()
            Patterns = ap.getPatterns()
            print("Total number of huis:", len(Patterns))
            ap.savePatterns('/home/apiiit-rkv/Downloads/output')
            memUSS = ap.getMemoryUSS()
            print("Total Memory in USS:", memUSS)
            memRSS = ap.getMemoryRSS()
            print("Total Memory in RSS", memRSS)
            run = ap.getRuntime()
            print("Total ExecutionTime in ms:", run)
        print("Error! The number of input parameters do not match the total number of parameters provided")



from PAMI.uncertainCorrelatedPattern.basic.abstract import *
import pandas as pd


class FFList:
    """
     A class represent a Fuzzy List of an element

    Attributes:
    ----------
         item: int
             the item name
         sumIUtil: float
             the sum of utilities of an fuzzy item in database
         sumRUtil: float
             the sum of resting values of a fuzzy item in database
         elements: list
             a list of elements contain tid, Utility and resting values of element in each transaction
    Methods:
    -------
        addElement(element)
            Method to add an element to this fuzzy list and update the sums at the same time.

        printElement(e)
            Method to print elements            

    """
    def __init__(self, itemName, region):
        self.item = itemName
        self.region = region
        self.sumIUtil = 0.0 
        self.sumRUtil = 0.0
        self.elements = []

    def addElement(self, element):
        self.sumIUtil += element.iUtils
        self.sumRUtil += element.rUtils
        self.elements.append(element) 


class Element:
    """
        A class represents an Element of a fuzzy list
    Attributes:
    ----------
        tid : int
            keep tact of transaction id
        iUtils: float
            the utility of an fuzzy item in the transaction
        rUtil : float
            the neighbourhood resting value of an fuzzy item in the transaction
    """
    def __init__(self, tid, iUtil, rUtil):
        self.tid = tid 
        self.iUtils = iUtil
        self.rUtils = rUtil


class Regions:
    """
        A class calculate the regions
        Attributes:
        ----------
            low : int
                low region value
            middle: int 
                middle region value
            high : int
                high region values
    """
    def __init__(self,  item,  quantity,  regionsNumber,  mapOfregios):
        self.low = 0
        self.middle = 0
        self.high = 0
        if regionsNumber == 3:
            if 0 < quantity <= 1:
                self.low = 1
                self.high = 0
                self.middle = 0
                t1 = (item,  'L')
                if t1 not in mapOfregios.keys():
                    mapOfregios[t1] = 1
                else:
                    temp = mapOfregios[t1]
                    mapOfregios[t1] = temp+1
            elif 1 <= quantity < 6:
                self.low = float((-0.2 * quantity) + 1.2)
                self.middle = float((0.2 * quantity) - 0.2)
                self.high = 0
                t1 = (item,  'L')
                if t1 not in mapOfregios.keys():
                    mapOfregios[t1] = 1
                else:
                    temp = mapOfregios[t1]
                    mapOfregios[t1] = temp+1
                t1 = (item,  'M')
                if t1 not in mapOfregios.keys():
                    mapOfregios[t1] = 1
                else:
                    temp = mapOfregios[t1]
                    mapOfregios[t1] = temp+1
            elif 6 <= quantity <= 11:
                self.low = 0
                self.middle = float((-0.2 * quantity) + 2.2)
                self.high = float((0.2 * quantity) - 1.2)
                t1 = (item,  'M')
                if t1 not in mapOfregios.keys():
                    mapOfregios[t1] = 1
                else:
                    temp = mapOfregios[t1]
                    mapOfregios[t1] = temp + 1
                t1 = (item,  'H')
                if t1 not in mapOfregios.keys():
                    mapOfregios[t1] = 1
                else:
                    temp = mapOfregios[t1]
                    mapOfregios[t1] = temp+1
            else:
                self.low = 0
                self.middle = 0
                self.high = 1
                t1 = (item,  'H')
                if t1 not in mapOfregios.keys():
                    mapOfregios[t1] = 1
                else:
                    temp = mapOfregios[t1]
                    mapOfregios = temp+1


class Pair:
    def __init__(self):
        """
            A Class to Store item and its quantity together
        """
        self.item = 0
        self.quantity = 0
        self.region = 'N'


class CFFI(frequentPatterns):
    """
        CFFI is the algorithm to discover Correlated Fuzzy-frequent patterns in a transactional database.
        it is based on traditional fuzzy frequent pattern mining.

    Attributes:
    ----------
        self.iFile : file
            Name of the input file to mine complete set of fuzzy spatial frequent patterns
        self. oFile : file
            Name of the oFile file to store complete set of fuzzy spatial frequent patterns
        minSup : int
            The user given support
        neighbors: map
            keep track of neighbours of elements
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program
        startTime:float
            To record the start time of the mining process
        endTime:float
            To record the completion time of the mining process
        itemsCnt: int
            To record the number of fuzzy spatial itemSets generated
        mapItemsLowSum: map
            To keep track of low region values of items
        mapItemsMidSum: map
            To keep track of middle region values of items
        mapItemsHighSum: map
            To keep track of high region values of items
        mapItemSum: map
            To keep track of sum of Fuzzy Values of items
        mapItemRegions: map
            To Keep track of fuzzy regions of item
        jointCnt: int
            To keep track of the number of FFI-list that was constructed
        BufferSize: int
            represent the size of Buffer
        itemBuffer list
            to keep track of items in buffer
    Methods:
    -------
        startMine()
            Mining process will start from here
        getFrequentPatterns()
            Complete set of patterns will be retrieved with this function
        savePatterns(oFile)
            Complete set of frequent patterns will be loaded in to a output file
        getPatternsAsDataFrame()
            Complete set of frequent patterns will be loaded in to a dataframe
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function            
        convert(value):
            To convert the given user specified value        

    Executing the code on terminal:
    -------
        Format: 
            python3 CFFI.py <inputFile> <outputFile> <minSup> <ratio>
        Examples: 
            python3 CFFI.py sampleTDB.txt output.txt 2 0.2 (minSup will be considered in support count or frequency)
                      
            python3 CFFI.py sampleTDB.txt output.txt 0.25 0.2 (minSup and maxPer will be considered in percentage of database)

    Sample run of importing the code:
    -------------------------------
        
        from PAMI.uncertainCorrelatedPattern import CFFI as alg

        obj = alg.CFFI("input.txt", 2, 0.4)

        obj.startMine()

        Patterns = obj.getPatterns()

        print("Total number of Correlated Fuzzy Frequent Patterns:",  len(Patterns))

        obj.savePatterns("outputFile")

        memUSS = obj.getMemoryUSS()

        print("Total Memory in USS:",  memUSS)

        memRSS = obj.getMemoryRSS()

        print("Total Memory in RSS",  memRSS)

        run = obj.getRuntime

        print("Total ExecutionTime in seconds:",  run)

    Credits:
    -------
        The complete program was written by Sai Chitra.B under the supervision of Professor Rage Uday Kiran.

    """
    startTime = float()
    endTime = float()
    minSup = str()
    maxPer = float()
    finalPatterns = {}
    iFile = " "
    oFile = " "
    sep = " "
    memoryUSS = float()
    memoryRSS = float()
    
    def __init__(self,  iFile,  minSup,  ratio,  sep):
        super().__init__(iFile,  minSup,  ratio)
        self.temp = {}
        self.mapItemRegionSum = {}
        self.start = 0
        self.end = 0
        self.itemsCnt = 0
        self.sep = sep
        self.mapItemsLowSum = {}
        self.mapItemsMidSum = {}
        self.mapItemsHighSum = {}            
        self.mapItemSum = {}
        self.mapItemRegions = {}
        self.joinsCnt = 0 
        self.BufferSize = 200
        self.itemSetBuffer = []
        self.finalPatterns = {}
        self.dbLen = 0

    def compareItems(self,  o1,  o2):
        """
            A Function that sort all FFI-list in ascending order of Support
        """
        compare = self.mapItemSum[o1.item] - self.mapItemSum[o2.item]
        if compare == 0:
            return o1.item - o2.item
        else:
            return compare

    @staticmethod
    def findElementWithTID(uList,  tid):
        """
        To find element with same tid as given
        :param uList: fuzzyList
        :type uList: FFI-List
        :param tid: transaction id
        :type tid: int
        :return: element eith tid as given
        :rtype: element if exists or None
        """
        List = uList.elements
        first = 0
        last = len(List)-1
        while first <= last:
            mid = (first + last) >> 1
            if List[mid].tid < tid:
                first = mid + 1
            elif List[mid].tid > tid:
                last = mid - 1
            else:
                return List[mid]
        return None            

    def convert(self,  value):
        """
        To convert the given user specified value
        :param value: user specified value
        :return: converted value
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self.Database) * value)
        if type(value) is str:
            if '.' in value:
                value = (len(self.Database) * value)
            else:
                value = int(value)
        return value
    
    def creatingItemSets(self):
        self.Database = []
        if isinstance(self.iFile, pd.DataFrame):
            if self.iFile.empty:
                print("its empty..")
            i = self.iFile.columns.values.tolist()
            if 'Transactions' in i:
                self.Database = self.iFile['Transactions'].tolist()
            if 'Patterns' in i:
                self.Database = self.iFile['Patterns'].tolist()
            # print(self.Database)
        if isinstance(self.iFile, str):
            if validators.url(self.iFile):
                data = urlopen(self.iFile)
                for line in data:
                    line = line.decode("utf-8")
                    self.Database.append(line)
            else:
                try:
                    with open(self.iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            self.Database.append(line)
                except IOError:
                    print("File Not Found")
                    quit()
                
    def startMine(self):
        """ 
            Frequent pattern mining process will start from here
        """
        self.startTime = time.time()
        self.creatingItemSets()
        self.finalPatterns = {}
        self.minSup = self.convert(self.minSup)
        minSup = self.minSup
        for line in self.Database:
            parts = line.split(":")
            items = parts[0].split(self.sep) 
            quantities = parts[1].split(self.sep) 
            for i in range(0, len(items)):
                item = items[i]
                regions = Regions(item, int(quantities[i]), 3, self.mapItemRegionSum)
                if item in self.mapItemsLowSum.keys():
                    low = self.mapItemsLowSum[item]
                    low += regions.low
                    self.mapItemsLowSum[item] = low
                else:
                    self.mapItemsLowSum[item] = regions.low
                if item in self.mapItemsMidSum.keys():
                    mid = self.mapItemsMidSum[item]
                    mid += regions.middle
                    self.mapItemsMidSum[item] = mid
                else:
                    self.mapItemsMidSum[item] = regions.middle
                if item in self.mapItemsHighSum.keys():
                    high = self.mapItemsHighSum[item]
                    high += regions.high
                    self.mapItemsHighSum[item] = high
                else:
                    self.mapItemsHighSum[item] = regions.high
        listOfFFIList = []
        mapItemsToFFLIST = {}
        for item1 in self.mapItemsLowSum.keys():
            item = item1
            region = 'N'
            low = self.mapItemsLowSum[item]
            mid = self.mapItemsMidSum[item]
            high = self.mapItemsHighSum[item]
            if low >= mid and low >= high:
                self.mapItemSum[item] = low
                self.mapItemRegions[item] = "L"
                region = 'L'
            elif mid >= low and mid >= high:
                self.mapItemSum[item] = mid
                self.mapItemRegions[item] = "M"
                region = 'M'
            elif high >= low and high >= mid:
                self.mapItemRegions[item] = "H"
                region = 'H'
                self.mapItemSum[item] = high
            if self.mapItemSum[item] >= self.minSup:
                fuList = FFList(item, region)
                mapItemsToFFLIST[item] = fuList 
                listOfFFIList.append(fuList)
        listOfFFIList.sort(key=functools.cmp_to_key(self.compareItems))
        tid = 0
        for line in self.Database:
            parts = line.split(":")
            items = parts[0].split(self.sep) 
            quantities = parts[1].split(self.sep) 
            revisedTransaction = []
            for i in range(0, len(items)):
                pair = Pair() 
                pair.item = items[i]
                regions = Regions(pair.item, int(quantities[i]), 3, self.temp)
                item = pair.item
                if self.mapItemSum[item] >= minSup:
                    if self.mapItemRegions[pair.item] == "L":
                        pair.quantity = regions.low
                        pair.region = 'L'
                    elif self.mapItemRegions[pair.item] == "M":
                        pair.region = 'M'
                        pair.quantity = regions.middle
                    elif self.mapItemRegions[pair.item] == "H":
                        pair.quantity = regions.high
                        pair.region = 'H'
                    if pair.quantity > 0:
                        revisedTransaction.append(pair)
            revisedTransaction.sort(key=functools.cmp_to_key(self.compareItems))
            for i in range(len(revisedTransaction)-1, -1, -1):
                pair = revisedTransaction[i]
                remainUtil = 0
                for j in range(len(revisedTransaction)-1, i-1, -1):
                    remainUtil += revisedTransaction[j].quantity
                if pair.quantity > remainUtil:
                    remainingUtility = pair.quantity
                else:
                    remainingUtility = remainUtil
                if mapItemsToFFLIST.get(pair.item) is not None:
                    FFListOfItem = mapItemsToFFLIST[pair.item]
                    element = Element(tid, pair.quantity, remainingUtility)
                    FFListOfItem.addElement(element)
            tid += 1
        self.FSFIMining(self.itemSetBuffer, 0, listOfFFIList, self.minSup) 
        self.endTime = time.time() 
        process = psutil.Process(os.getpid())
        self.memoryUSS = float()
        self.memoryRSS = float()
        self.memoryUSS = process.memory_full_info().uss
        self.memoryRSS = process.memory_info().rss
        
    def FSFIMining(self, prefix, prefixLen, FSFIM, minSup):
        """
            Generates FFSI from prefix

        :param prefix: the prefix patterns of FFSI
        :type prefix: len
        :param prefixLen: the length of prefix
        :type prefixLen: int
           :param FSFIM: the Fuzzy list of prefix itemSets
           :type FSFIM: list
           :param minSup: the minimum support of 
           :type minSup:int
        """
        for i in range(0, len(FSFIM)):
            X = FSFIM[i]
            if X.sumIUtil >= minSup:
                ratio = self.getRatio(prefix, prefixLen, X)
                if ratio >= self.minRatio:
                    self.WriteOut(prefix, prefixLen, X, ratio) 
            if X.sumRUtill >= minSup:
                exULs = []
                for j in range(i+1, len(FSFIM)):
                    Y = FSFIM[j]
                    exULs.append(self.construct(X, Y))
                    self.joinsCnt += 1
                self.itemSetBuffer.insert(prefixLen, X)
                self.FSFIMining(self.itemSetBuffer, prefixLen+1, exULs, minSup)
    
    def construct(self, px, py):
        """
            A function to construct a new Fuzzy itemSet from 2 fuzzy itemSets

            :param px:the itemSet px
            :type px:FFI-List
            :param py:itemSet py
            :type py:FFI-List
            :return :the itemSet of pxy(px and py)
            :rtype :FFI-List
        """
        pxyUL = FFList(py.item, py.region)
        for ex in px.elements:
            ey = self.findElementWithTID(py,  ex.tid)
            if ey is None:
                continue
            eXY = Element(ex.tid, min([ex.iUtils, ey.iUtils], key=lambda x: float(x)), ey.rUtils)
            pxyUL.addElement(eXY)
        return pxyUL

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self.memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
       """
        return self.memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process


        :return: returning total amount of runtime taken by the mining process
        :rtype: float
       """
        return self.endTime-self.start
    
    def getRatio(self, prefix, prefixLen, item):
        """Method to calculate the ration of itemSet
            :param prefix: prefix of itemSet
            :type prefix: list
            :param prefixLen: length of prefix
            :type prefixLen: int
            :param item: the last item
            :type item: Node
            :return : correlatedSpatialPattern ratio
            :rtype: float
        """
        res = 1.0
        for i in prefix:
            if self.mapItemRegionSum.get((i.item,  i.region)) is not None and res < self.mapItemRegionSum[(i.item,  i.region)]:
                res = self.mapItemRegionSum[(i.item, i.region)]
        if self.mapItemRegionSum.get((item.item,  item.region)) is not None and res < self.mapItemRegionSum[(item.item,  item.region)]:
            res = self.mapItemRegionSum[(item.item, item.region)]
        return item.sumIUtil / res

    def WriteOut(self, prefix, prefixLen, item, ratio):
        """
            To Store the patten
            :param prefix: prefix of itemSet
            :type prefix: list
            :param prefixLen: length of prefix
            :type prefixLen: int
            :param item: the last item
            :type item: Node
            :param ratio: the ratio of itemSet
            :type ratio: float
        """
        self.itemsCnt += 1
        res = "" 
        for i in range(0, prefixLen):
            res += str(prefix[i].item)+"."+str(prefix[i].region)+' '
        res += str(item.item)+"."+str(item.region) 
        res1 = str(item.sumIUtil)+" : "+str(ratio)+"\n"
        self.finalPatterns[res] = res1 

    def getFrequentPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self.finalPatterns
    
    def getPatternsAsDataFrame(self):
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataframe = {}
        data = []
        for a,  b in self.finalPatterns.items():
            data.append([a,  b])
            dataframe = pd.DataFrame(data,  columns=['Patterns',  'Support'])
        return dataframe

    def savePatterns(self,  outFile):
        """Complete set of frequent patterns will be loaded in to a output file

        :param outFile: name of the output file
        :type outFile: file
        """
        self.oFile = outFile
        writer = open(self.oFile,  'w+')
        for x,  y in self.finalPatterns.items():
            patternsAndSupport = str(x) + " : " + str(y)
            writer.write("%s \n" % patternsAndSupport)
    
    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self.finalPatterns


if __name__ == "__main__":
    ap = str()
    if len(sys.argv) == 5 or len(sys.argv) == 6:
        if len(sys.argv) == 6:
            ap = CFFI(sys.argv[1], sys.argv[3], sys.argv[4], sys.argv[5])
        if len(sys.argv) == 5:
            ap = CFFI(sys.argv[1], sys.argv[3], sys.argv[4])
        ap.startMine()
        Patterns = ap.getPatterns()
        print("Total number of Frequent Patterns:", len(Patterns))
        ap.savePatterns(sys.argv[2])
        memUSS = ap.getMemoryUSS()
        print("Total Memory in USS:", memUSS)
        memRSS = ap.getMemoryRSS()
        print("Total Memory in RSS", memRSS)
        run = ap.getRuntime()
        print("Total ExecutionTime in ms:", run)
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")
        



import sys
import validators
from urllib.request import urlopen
from PAMI.partialPeriodicPattern.closed.abstract import *


class PPPClose(partialPeriodicPatterns):
    """ PPPClose algorithm is used to discover the closed partial periodic patterns in temporal databases.
        It uses depth-first search.

        Reference:
        -------
        ...
        
        Attributes:
        ----------
            iFile : str
                Input file name or path of the input file
            oFile : str
                Name of the output file or path of the input file
            periodicSupport: int or float or str
                The user can specify periodicSupport either in count or proportion of database size.
                If the program detects the data type of periodicSupport is integer, then it treats periodicSupport is expressed in count.
                Otherwise, it will be treated as float.
                Example: periodicSupport=10 will be treated as integer, while periodicSupport=10.0 will be treated as float
            period: int or float or str
                The user can specify period either in count or proportion of database size.
                If the program detects the data type of period is integer, then it treats period is expressed in count.
                Otherwise, it will be treated as float.
                Example: period=10 will be treated as integer, while period=10.0 will be treated as float
            sep : str
                This variable is used to distinguish items from one another in a transaction. The default separator is tab space or \t.
                However, the users can override their default separator.
            startTime:float
                To record the start time of the mining process
            endTime:float
                To record the completion time of the mining process
            finalPatterns: dict
                Storing the complete set of patterns in a dictionary variable
            memoryUSS : float
                To store the total amount of USS memory consumed by the program
            memoryRSS : float
                To store the total amount of RSS memory consumed by the program

        Methods:
        -------

            startMine()
                Mining process will start from here
            getPatterns()
                Complete set of patterns will be retrieved with this function
            savePatterns(oFile)
                Complete set of frequent patterns will be loaded in to a output file
            getPatternsAsDataFrame()
                Complete set of frequent patterns will be loaded in to a dataframe
            getMemoryUSS()
                Total amount of USS memory consumed by the mining process will be retrieved from this function
            getMemoryRSS()
                Total amount of RSS memory consumed by the mining process will be retrieved from this function
            getRuntime()
                Total amount of runtime taken by the mining process will be retrieved from this function

        Executing the code on terminal:
        -------
        Format:
        ------
            python3 PPPClose.py <inputFile> <outputFile> <periodicSupport> <period>

        Examples:
        --------
            python3 PPPClose.py sampleTDB.txt patterns.txt 0.3 0.4   (periodicSupport and period will be considered in percentage of database
        transactions)

            python3 PPPClose.py sampleTDB.txt patterns.txt 3 4     (periodicSupport and period will be considered in support count or frequency)


        Sample run of the imported code:
        --------------

            from PAMI.partialPeriodicPattern.closed import PPPClose as alg

            obj = alg.PPPClose("../basic/sampleTDB.txt", "2", "6")

            obj.startMine()

            periodicFrequentPatterns = obj.getPatterns()

            print("Total number of Frequent Patterns:", len(periodicFrequentPatterns))

            obj.savePatterns("patterns")

            Df = obj.getPatternsAsDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

        Credits:
        -------
            The complete program was written by P.Likhitha  under the supervision of Professor Rage Uday Kiran.\n

        """

    periodicSupport = float()
    period = float()
    startTime = float()
    endTime = float()
    finalPatterns = {}
    Database = []
    iFile = " "
    oFile = " "
    sep = " "
    memoryUSS = float()
    memoryRSS = float()
    transaction = []
    hashing = {}
    mapSupport = {}
    itemSetCount = 0
    maxItemId = 0
    tableSize = 10000
    tidList = {}
    lno = 0

    def convert(self, value):
        """
        To convert the given user specified value

        :param value: user specified value

        :return: converted value
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (self.lno * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (self.lno * value)
            else:
                value = int(value)
        return value
    
    def creatingItemSets(self):
        """
            Storing the complete transactions of the database/input file in a database variable


        """
        self.Database = []
        if isinstance(self.iFile, pd.DataFrame):
            timeStamp, data = [], []
            if self.iFile.empty:
                print("its empty..")
            i = self.iFile.columns.values.tolist()
            if 'TS' in i:
                timeStamp = self.iFile['TS'].tolist()
            if 'Transactions' in i:
                data = self.iFile['Transactions'].tolist()
            for i in range(len(data)):
                tr = [timeStamp[i]]
                tr = tr + data[i]
                self.Database.append(tr)
            self.lno = len(self.Database)
        if isinstance(self.iFile, str):
            if validators.url(self.iFile):
                data = urlopen(self.iFile)
                for line in data:
                    self.lno += 1
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self.sep)]
                    temp = [x for x in temp if x]
                    self.Database.append(temp)
            else:
                try:
                    with open(self.iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            self.lno += 1
                            temp = [i.rstrip() for i in line.split(self.sep)]
                            temp = [x for x in temp if x]
                            self.Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()

    def OneLengthPartialItems(self):
        """
        To scan the database and extracts the 1-length periodic-frequent items
        Returns:
        -------
        Returns the 1-length periodic-frequent items
        """
        self.mapSupport = {}
        self.tidList = {}
        self.period = self.convert(self.period)
        for line in self.Database:
            n = int(line[0])
            for i in range(1, len(line)):
                si = line[i]
                if self.mapSupport.get(si) is None:
                    self.mapSupport[si] = [1, 0, n]
                    self.tidList[si] = [n]
                else:
                    self.mapSupport[si][0] += 1
                    period = abs(n - self.mapSupport[si][2])
                    if period <= self.period:
                        self.mapSupport[si][1] += 1
                    self.mapSupport[si][2] = n
                    self.tidList[si].append(n)
        for x, y in self.mapSupport.items():
            period = abs(self.lno - self.mapSupport[x][2])
            if period <= self.period:
                self.mapSupport[x][1] += 1
        self.periodicSupport = self.convert(self.periodicSupport)
        self.mapSupport = {k: v[1] for k, v in self.mapSupport.items() if v[1] >= self.periodicSupport}
        periodicFrequentItems = {}
        self.tidList = {k: v for k, v in self.tidList.items() if k in self.mapSupport}
        for x, y in self.tidList.items():
            t1 = 0
            for i in y:
                t1 += i
            periodicFrequentItems[x] = t1
        periodicFrequentItems = [key for key, value in sorted(periodicFrequentItems.items(), key=lambda x: x[1])]
        return periodicFrequentItems

    def calculate(self, tidSet):
        """
        To calculate the weight if pattern based on the respective timeStamps
        Parameters
        ----------
        tidSet: timeStamps of the pattern

        Returns
        -------
        the calculated weight of the timeStamps
        """
        hashcode = 0
        for i in tidSet:
            hashcode += i
        if hashcode < 0:
            hashcode = abs(0 - hashcode)
        return hashcode % self.tableSize

    def contains(self, itemSet, val, hashcode):
        """
        To check if the key(hashcode) is in dictionary(hashing) variable
        Parameters:
        ----------
            itemSet: generated periodic-frequent itemSet
            val: support and period of itemSet
            hashcode: the key generated in calculate() method for every itemSet

        Returns
        -------
            true if itemSet with same support present in dictionary(hashing) or else returns false
        """
        if self.hashing.get(hashcode) is None:
            return False
        for i in self.hashing[hashcode]:
            itemSetX = i
            if val == self.hashing[hashcode][itemSetX] and set(itemSetX).issuperset(itemSet):
                return True
        return False

    def getPeriodicSupport(self, timeStamps):
        """
        Calculates the period and support of timeStamps
        Parameters:
        ----------
            timeStamps: timeStamps of itemSet

        Returns:
        -------
            period and support
        """
        timeStamps.sort()
        sup = 0
        for j in range(len(timeStamps) - 1):
            per = abs(timeStamps[j + 1] - timeStamps[j])
            if per <= self.period:
                sup += 1
        return sup

    def save(self, prefix, suffix, tidSetX):
        """
        Saves the generated pattern which satisfies the closed property
        Parameters:
        ----------
            prefix: the prefix part of itemSet
            suffix: the suffix part of itemSet
            tidSetX: the timeStamps of the generated itemSet

        Returns:
        -------
            saves the closed periodic-frequent pattern

        """
        if prefix is None:
            prefix = suffix
        else:
            prefix = prefix + suffix
        prefix = list(set(prefix))
        prefix.sort()
        val = self.getPeriodicSupport(tidSetX)
        if val >= self.periodicSupport:
            hashcode = self.calculate(tidSetX)
            if self.contains(prefix, val, hashcode) is False:
                self.itemSetCount += 1
                sample = str()
                for i in prefix:
                    sample = sample + i + " "
                self.finalPatterns[sample] = val
            if hashcode not in self.hashing:
                self.hashing[hashcode] = {tuple(prefix): val}
            else:
                self.hashing[hashcode][tuple(prefix)] = val

    def processEquivalenceClass(self, prefix, itemSets, tidSets):
        """
        Parameters:
        ----------
            prefix: Prefix class of an itemSet
            itemSets: suffix items in periodicFrequentItems that satisfies the periodicSupport condition
            tidSets: timeStamps of items in itemSets respectively

        Returns:
        -------
            closed periodic patterns with length more than 2
        """
        if len(itemSets) == 1:
            i = itemSets[0]
            tidList = tidSets[0]
            self.save(prefix, [i], tidList)
            return
        if len(itemSets) == 2:
            itemI = itemSets[0]
            tidSetI = tidSets[0]
            itemJ = itemSets[1]
            tidSetJ = tidSets[1]
            y1 = list(set(tidSetI).intersection(tidSetJ))
            if len(y1) >= self.periodicSupport:
                suffix = []
                suffix += [itemI, itemJ]
                suffix = list(set(suffix))
                self.save(prefix, suffix, y1)
            if len(y1) != len(tidSetI):
                self.save(prefix, [itemI], tidSetI)
            if len(y1) != len(tidSetJ):
                self.save(prefix, [itemJ], tidSetJ)
            return
        for i in range(len(itemSets)):
            itemX = itemSets[i]
            if itemX is None:
                continue
            tidSetX = tidSets[i]
            classItemSets = []
            classTidSets = []
            itemSetX = [itemX]
            for j in range(i + 1, len(itemSets)):
                itemJ = itemSets[j]
                if itemJ is None:
                    continue
                tidSetJ = tidSets[j]
                y = list(set(tidSetX).intersection(tidSetJ))
                if len(y) < self.periodicSupport:
                    continue
                if len(tidSetX) == len(tidSetJ) and len(y) == len(tidSetX):
                    itemSets.insert(j, None)
                    tidSets.insert(j, None)
                    itemSetX.append(itemJ)
                elif len(tidSetX) < len(tidSetJ) and len(y) == len(tidSetX):
                    itemSetX.append(itemJ)
                elif len(tidSetX) > len(tidSetJ) and len(y) == len(tidSetJ):
                    itemSets.insert(j, None)
                    tidSets.insert(j, None)
                    classItemSets.append(itemJ)
                    classTidSets.append(y)
                else:
                    classItemSets.append(itemJ)
                    classTidSets.append(y)
            if len(classItemSets) > 0:
                newPrefix = list(set(itemSetX)) + prefix
                self.processEquivalenceClass(newPrefix, classItemSets, classTidSets)
            self.save(prefix, list(set(itemSetX)), tidSetX)

    def startMine(self):
        """
        Mining process will start from here
        """
        self.startTime = time.time()
        self.creatingItemSets()
        self.finalPatterns = {}
        periodicFrequentItems = self.OneLengthPartialItems()
        for i in range(len(periodicFrequentItems)):
            itemX = periodicFrequentItems[i]
            if itemX is None:
                continue
            tidSetX = self.tidList[itemX]
            itemSetX = [itemX]
            itemSets = []
            tidSets = []
            for j in range(i + 1, len(periodicFrequentItems)):
                itemJ = periodicFrequentItems[j]
                if itemJ is None:
                    continue
                tidSetJ = self.tidList[itemJ]
                y1 = list(set(tidSetX).intersection(tidSetJ))
                if len(y1) < self.periodicSupport:
                    continue
                if len(tidSetX) == len(tidSetJ) and len(y1) is len(tidSetX):
                    periodicFrequentItems.insert(j, None)
                    itemSetX.append(itemJ)
                elif len(tidSetX) < len(tidSetJ) and len(y1) is len(tidSetX):
                    itemSetX.append(itemJ)
                elif len(tidSetX) > len(tidSetJ) and len(y1) is len(tidSetJ):
                    periodicFrequentItems.insert(j, None)
                    itemSets.append(itemJ)
                    tidSets.append(y1)
                else:
                    itemSets.append(itemJ)
                    tidSets.append(y1)
            if len(itemSets) > 0:
                self.processEquivalenceClass(itemSetX, itemSets, tidSets)
            self.save([], itemSetX, tidSetX)
        self.endTime = time.time()
        process = psutil.Process(os.getpid())
        self.memoryUSS = float()
        self.memoryRSS = float()
        self.memoryUSS = process.memory_full_info().uss
        self.memoryRSS = process.memory_info().rss
        print("Closed periodic frequent patterns were generated successfully using PPPClose algorithm ")

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

            :return: returning USS memory consumed by the mining process

            :rtype: float
        """

        return self.memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

            :return: returning RSS memory consumed by the mining process

            :rtype: float
        """

        return self.memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process

            :return: returning total amount of runtime taken by the mining process

            :rtype: float
        """

        return self.endTime - self.startTime

    def getPatternsAsDataFrame(self):
        """Storing final frequent patterns in a dataframe

            :return: returning frequent patterns in a dataframe

            :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self.finalPatterns.items():
            data.append([a, b[0], b[1]])
            dataFrame = pd.DataFrame(data, columns=['Patterns', 'Support', 'Period'])
        return dataFrame

    def savePatterns(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file

            :param outFile: name of the output file

            :type outFile: file
        """
        self.oFile = outFile
        writer = open(self.oFile, 'w+')
        for x, y in self.finalPatterns.items():
            s1 = x + ":" + str(y[0]) + ":" + str(y[1])
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

            :return: returning frequent patterns

            :rtype: dict
        """
        return self.finalPatterns


if __name__ == "__main__":
    ap = str()
    if len(sys.argv) == 5 or len(sys.argv) == 6:
        if len(sys.argv) == 6:
            ap = PPPClose(sys.argv[1], sys.argv[3], sys.argv[4], sys.argv[5])
        if len(sys.argv) == 5:
            ap = PPPClose(sys.argv[1], sys.argv[3], sys.argv[4])
        ap.startMine()
        Patterns = ap.getPatterns()
        print("Total number of  Patterns:", len(Patterns))
        ap.savePatterns(sys.argv[2])
        memUSS = ap.getMemoryUSS()
        print("Total Memory in USS:", memUSS)
        memRSS = ap.getMemoryRSS()
        print("Total Memory in RSS", memRSS)
        run = ap.getRuntime()
        print("Total ExecutionTime in ms:", run)
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")





class RuleMiner:
    """
    temporalDatabaseStats is class to get stats of database.

        Attributes:
        ----------
        frequentPattern : list or dict
            list
        measure: condition to calculate the strength of rule
            str
        threshold: condition to satisfy
            int

        Methods:
        -------
        startMine()

    """

    def __init__(self, frequentPatterns, measure, threshold):
        """
        :param inputFile: input file name or path
        :type inputFile: str
        :param sep:
        """
        self.frequentPattern = frequentPatterns
        self.measure = measure
        self.threshold = threshold

    def save(self, prefix, suffix):
        print(prefix, suffix)

    def generation(self, prefix, suffix):
        if len(suffix) == 0:
            return
        for i in range(len(suffix)):
            prefix = prefix + suffix[i]
            sample = str()
            for j in suffix[i+1:]:
                sample = sample + j
                print(prefix, sample)
                self.generation(prefix, sample)

    def RuleMiner(self):
        flist = []
        for i in self.frequentPattern.keys():
            for j in i:
                if j not in flist:
                    flist.append(j)
        for i in range(len(flist)):
            sample = str()
            for j in flist[i+1:]:
                sample = sample + j
                print(flist[i], sample)
                self.generation(flist[i], sample)



if __name__ == '__main__':
    data = {'c':2, 'cb':2, 'b':5,
            'bd':3, 'ba':3, 'd':5, 'da':3, 'a':5}
    ap = RuleMiner(data, "confidence", 50)
    ap.RuleMiner()


import folium
import pandas as pd

def plotPointIntoMap(inputFile):
    mmap = folium.Map(location=[35.39, 139.44], zoom_start=5)
    df = pd.read_csv(inputFile)
    colors = ['red', 'blue', 'green', 'purple', 'orange', 'darkred', 'lightred', 'beige', 'darkblue', 'darkgreen',
              'cadetblue', 'darkpurple', 'white', 'pink', 'lightblue', 'lightgreen', 'gray', 'black', 'lightgray']
    for i, row in df.iterrows():
        folium.CircleMarker(
            location=(row['latitude'], row['longitude']),
            popup=row['patternId'],
            radius=3,
            # icon=folium.Icon(color=colors[int(row['patternId'])-1])
            color=colors[int(row['patternId']) - 1],
            fill=True,
            fill_color=colors[int(row['patternId']) - 1],
        ).add_to(mmap)
    return mmap


#  Copyright (C)  2021 Rage Uday Kiran
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
# from abstract import *

from PAMI.frequentPattern.basic.abstract import *


class ECLAT(frequentPatterns):
    """ ECLAT is one of the fundamental algorithm to discover frequent patterns in a transactional database.
        This program employs downward closure property to  reduce the search space effectively.
        This algorithm employs depth-first search technique to find the complete set of frequent patterns in a
        transactional database.

        Reference:
        ----------
            Mohammed Javeed Zaki: Scalable Algorithms for Association Mining. IEEE Trans. Knowl. Data Eng. 12(3):
            372-390 (2000), https://ieeexplore.ieee.org/document/846291

        Attributes:
        ----------
            iFile : str
                Input file name or path of the input file
            minSup: float or int or str
                The user can specify minSup either in count or proportion of database size.
                If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.
                Otherwise, it will be treated as float.
                Example: minSup=10 will be treated as integer, while minSup=10.0 will be treated as float
            sep : str
                This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
                However, the users can override their default separator.
            oFile : str
                Name of the output file or the path of the output file
            startTime:float
                To record the start time of the mining process
            endTime:float
                To record the completion time of the mining process
            finalPatterns: dict
                Storing the complete set of patterns in a dictionary variable
            memoryUSS : float
                To store the total amount of USS memory consumed by the program
            memoryRSS : float
                To store the total amount of RSS memory consumed by the program

        Methods:
        -------
            startMine()
                Mining process will start from here
            getPatterns()
                Complete set of patterns will be retrieved with this function
            savePatterns(oFile)
                Complete set of frequent patterns will be loaded in to a output file
            getPatternsAsDataFrame()
                Complete set of frequent patterns will be loaded in to a dataframe
            getMemoryUSS()
                Total amount of USS memory consumed by the mining process will be retrieved from this function
            getMemoryRSS()
                Total amount of RSS memory consumed by the mining process will be retrieved from this function
            getRuntime()
                Total amount of runtime taken by the mining process will be retrieved from this function
            creatingItemSets()
                Scans the dataset or dataframes and stores in list format
            frequentOneItem()
                Generates one frequent patterns
            ECLATGeneration(candidateList)
                It will generate the combinations of frequent items
            generateFrequentPatterns(tidList)
                It will generate the combinations of frequent items from a list of items

        Executing the code on terminal:
        -------------------------------

            Format:
            ------
            python3 ECLAT.py <inputFile> <outputFile> <minSup>

            Examples:
            ---------
            python3 ECLAT.py sampleDB.txt patterns.txt 10.0   (minSup will be considered in times of minSup and count of database transactions)

            python3 ECLAT.py sampleDB.txt patterns.txt 10     (minSup will be considered in support count or frequency)


        Sample run of the importing code:
        ---------------------------------

            import PAMI.frequentPattern.basic.ECLAT as alg

            obj = alg.ECLAT(iFile, minSup)

            obj.startMine()

            frequentPatterns = obj.getPatterns()

            print("Total number of Frequent Patterns:", len(frequentPatterns))

            obj.savePatterns(oFile)

            Df = obj.getPatternInDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

        Credits:
        --------
            The complete program was written by Kundai  under the supervision of Professor Rage Uday Kiran.

    """

    minSup = float()
    startTime = float()
    endTime = float()
    finalPatterns = {}
    iFile = " "
    oFile = " "
    sep = " "
    memoryUSS = float()
    memoryRSS = float()
    Database = []

    def creatingItemSets(self):
        """
            Storing the complete transactions of the database/input file in a database variable

        """
        self.Database = []
        if isinstance(self.iFile, pd.DataFrame):
            if self.iFile.empty:
                print("its empty..")
            i = self.iFile.columns.values.tolist()
            if 'Transactions' in i:
                self.Database = self.iFile['Transactions'].tolist()
        if isinstance(self.iFile, str):
            if validators.url(self.iFile):
                data = urlopen(self.iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self.sep)]
                    temp = [x for x in temp if x]
                    self.Database.append(temp)
            else:
                try:
                    with open(self.iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self.sep)]
                            temp = [x for x in temp if x]
                            self.Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()

    def getUniqueItemList(self):
        """
        Generating one frequent patterns
        """
        self.finalPatterns = {}
        candidate = {}
        uniqueItem = []
        for i in range(len(self.Database)):
            for j in range(len(self.Database[i])):
                if self.Database[i][j] not in candidate:
                    candidate[self.Database[i][j]] = {i}
                else:
                    candidate[self.Database[i][j]].add(i)
        for key, value in candidate.items():
            supp = len(value)
            if supp >= self.minSup:
                self.finalPatterns[key] = [value]
                uniqueItem.append(key)
        uniqueItem.sort(key=int)
        return uniqueItem

    def generateFrequentPatterns(self, candidateFrequent):
        """It will generate the combinations of frequent items

        :param candidateFrequent :it represents the items with their respective transaction identifiers

        :type candidateFrequent: list

        :return: returning transaction dictionary

        :rtype: dict
        """
        new_freqList = []
        for i in range(0, len(candidateFrequent)):
            item1 = candidateFrequent[i]
            i1_list = item1.split()
            for j in range(i + 1, len(candidateFrequent)):
                item2 = candidateFrequent[j]
                i2_list = item2.split()
                if i1_list[:-1] == i2_list[:-1]:
                    interSet = self.finalPatterns[item1][0].intersection(self.finalPatterns[item2][0])
                    if len(interSet) >= self.minSup:
                        newKey = item1 + " " + i2_list[-1]
                        self.finalPatterns[newKey] = [interSet]
                        new_freqList.append(newKey)

                    if len(new_freqList) > 0:
                        self.generateFrequentPatterns(new_freqList)

    def convert(self, value):
        """
        To convert the user specified minSup value

        :param value: user specified minSup value

        :return: converted type
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self.Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self.Database) * value)
            else:
                value = int(value)
        return value

    def startMine(self):
        """Frequent pattern mining process will start from here"""

        self.startTime = time.time()
        if self.iFile is None:
            raise Exception("Please enter the file path or file name:")
        if self.minSup is None:
            raise Exception("Please enter the Minimum Support")
        self.creatingItemSets()
        self.minSup = self.convert(self.minSup)
        uniqueItemList = self.getUniqueItemList()
        self.generateFrequentPatterns(uniqueItemList)
        for x, y in self.finalPatterns.items():
            self.finalPatterns[x] = len(y[0])
        self.endTime = time.time()
        process = psutil.Process(os.getpid())
        self.memoryUSS = float()
        self.memoryRSS = float()
        self.memoryUSS = process.memory_full_info().uss
        self.memoryRSS = process.memory_info().rss
        print("Frequent patterns were generated successfully using ECLAT algorithm")

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process

        :rtype: float
        """

        return self.memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process

        :rtype: float
        """

        return self.memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process

        :rtype: float
        """

        return self.endTime - self.startTime

    def getPatternsAsDataFrame(self):
        """Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe

        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self.finalPatterns.items():
            data.append([a, b])
            dataFrame = pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataFrame

    def savePatterns(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file

        :param outFile: name of the output file

        :type outFile: file
        """
        self.oFile = outFile
        writer = open(self.oFile, 'w+')
        for x, y in self.finalPatterns.items():
            patternsAndSupport = x + ":" + str(y)
            writer.write("%s \n" % patternsAndSupport)

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns

        :rtype: dict
        """
        return self.finalPatterns


if __name__ == "__main__":
    ap = str()
    if len(sys.argv) == 4 or len(sys.argv) == 5:
        if len(sys.argv) == 5:
            ap = ECLAT(sys.argv[1], sys.argv[3], sys.argv[4])
        if len(sys.argv) == 4:
            ap = ECLAT(sys.argv[1], sys.argv[3])
        ap.startMine()
        Patterns = ap.getPatterns()
        print("Total number of Frequent Patterns:", len(Patterns))
        ap.savePatterns(sys.argv[2])
        print(ap.getPatternsAsDataFrame())
        memUSS = ap.getMemoryUSS()
        print("Total Memory in USS:", memUSS)
        memRSS = ap.getMemoryRSS()
        print("Total Memory in RSS", memRSS)
        run = ap.getRuntime()
        print("Total ExecutionTime in ms:", run)
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")

        



from PAMI.periodicFrequentSpatialPattern.abstract import *
import sys
import validators
from urllib.request import urlopen
import pandas as pd


class PFS_ECLAT(spatialPeriodicFrequentPatterns):
    """ 
        Spatial Eclat is a Extension of ECLAT algorithm,which  stands for Equivalence Class Clustering and bottom-up
        Lattice Traversal.It is one of the popular methods of Association Rule mining. It is a more efficient and
        scalable version of the Apriori algorithm.
            ...
    Attributes :
    ----------
            iFile : str
                Input file name or path of the input file
            nFile: str:
               Name of Neighbourhood file name
            minSup: float or int or str
                The user can specify minSup either in count or proportion of database size.
                If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.
                Otherwise, it will be treated as float.
                Example: minSup=10 will be treated as integer, while minSup=10.0 will be treated as float
            maxPer: float or int or str
                The user can specify maxPer either in count or proportion of database size.
                If the program detects the data type of maxPer is integer, then it treats minSup is expressed in count.
                Otherwise, it will be treated as float.
                Example: maxPer=10 will be treated as integer, while maxPer=10.0 will be treated as float
            sep : str
                This variable is used to distinguish items from one another in a transaction. The default separator is tab space or \t.
                However, the users can override their default separator.
            startTime:float
                To record the start time of the mining process
            endTime:float
                To record the completion time of the mining process
            finalPatterns: dict
                Storing the complete set of patterns in a dictionary variable
            oFile : str
                Name of the output file to store complete set of frequent patterns
            memoryUSS : float
                To store the total amount of USS memory consumed by the program
            memoryRSS : float
                To store the total amount of RSS memory consumed by the program
            Database : list
                To store the complete set of transactions available in the input database/file
    Methods :
    -------
            startMine()
                Mining process will start from here
            getPatterns()
                Complete set of patterns will be retrieved with this function
            savePatterns(oFile)
                Complete set of frequent patterns will be loaded in to a output file
            getPatternsAsDataFrames()
                Complete set of frequent patterns will be loaded in to a dataframe
            getMemoryUSS()
                Total amount of USS memory consumed by the mining process will be retrieved from this function
            getMemoryRSS()
                Total amount of RSS memory consumed by the mining process will be retrieved from this function
            getRuntime()
                Total amount of runtime taken by the mining process will be retrieved from this function
            creatingItemSets(iFileName)
                Storing the complete transactions of the database/input file in a database variable
            frequentOneItem()
                Generating one frequent patterns
            convert(value):
                To convert the given user specified value    
            getNeighbourItems(keySet):
                A function to get common neighbours of a itemSet
             mapNeighbours(file):
                A function to map items to their neighbours
    Executing the code on terminal :
    ------------------------------
        Format:
            python3 PFS_ECLAT.py <inputFile> <outputFile> <neighbourFile> <minSup> <maxPer>
        Examples:
            python3 PFS_ECLAT.py sampleTDB.txt output.txt sampleN.txt 0.5 0.3 (minSup & maxPer will be considered in percentage of database transactions)

            python3 PFS_ECLAT.py sampleTDB.txt output.txt sampleN.txt 5 3 (minSup & maxPer will be considered in support count or frequency)
                                                                (it considers "\t" as separator)

            python3 PFS_ECLAT.py sampleTDB.txt output.txt sampleN.txt 3 ',' (it will consider "," as a separator)

    Sample run of importing the code :
    -------------------------------

        import PAMI.periodicFrequentSpatialPattern.periodicSpatialEclat as alg

        obj = alg.periodicSpatialEclat("sampleTDB.txt", "sampleN.txt", 5, 3)

        obj.startMine()

        spatialPeriodicFrequentPatterns = obj.getPatterns()

        print("Total number of Periodic Spatial Frequent Patterns:", len(spatialPeriodicFrequentPatterns))

        obj.savePatterns("outFile")

        memUSS = obj.getMemoryUSS()

        print("Total Memory in USS:", memUSS)

        memRSS = obj.getMemoryRSS()

        print("Total Memory in RSS", memRSS)

        run = obj.getRuntime()

        print("Total ExecutionTime in seconds:", run)

    Credits:
    -------
        The complete program was written by P. Likhitha under the supervision of Professor Rage Uday Kiran.
    """

    minSup = " "
    maxPer = " "
    startTime = float()
    endTime = float()
    finalPatterns = {}
    iFile = " "
    oFile = " "
    nFile = " "
    memoryUSS = float()
    memoryRSS = float()
    Database = []
    sep = "\t"
    lno = 0

    def __init__(self, iFile, nFile, minSup, maxPer, sep="\t"):
        super().__init__(iFile, nFile, minSup, maxPer, sep)
        self.NeighboursMap = {}

    def creatingItemSets(self, iFileName):
        """Storing the complete transactions of the database/input file in a database variable
            :param iFileName: user given input file/input file path
            :type iFileName: str
            """
        self.Database = []
        if isinstance(self.iFile, pd.DataFrame):
            data, ts = [], []
            if self.iFile.empty:
                print("its empty..")
            i = self.iFile.columns.values.tolist()
            if 'TS' in i:
                ts = self.iFile['TS'].tolist()
            if 'Transactions' in i:
                data = self.iFile['Transactions'].tolist()
            for i in range(len(data)):
                tr = [ts[i][0]]
                tr = tr + data[i]
                self.Database.append(tr)
        if isinstance(self.iFile, str):
            if validators.url(self.iFile):
                data = urlopen(self.iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self.sep)]
                    temp = [x for x in temp if x]
                    self.Database.append(temp)
            else:
                try:
                    with open(self.iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self.sep)]
                            temp = [x for x in temp if x]
                            self.Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()

    # function to get frequent one pattern
    def frequentOneItem(self):
        """Generating one frequent patterns"""

        candidate = {}
        for i in self.Database:
            self.lno += 1
            for j in i:
                if j not in candidate:
                    candidate[j] = [1, abs(0-self.lno), self.lno, [self.lno]]
                else:
                    candidate[j][0] += 1
                    candidate[j][1] = max(candidate[j][1], abs(self.lno - candidate[j][2]))
                    candidate[j][2] = self.lno
                    candidate[j][3].append(self.lno)
        self.minSup = self.convert(self.minSup)
        self.maxPer = self.convert(self.maxPer)
        self.tidList = {k: v[3] for k, v in candidate.items() if v[0] >= self.minSup and v[1] <= self.maxPer}
        candidate = {k: [v[0], v[1]] for k, v in candidate.items() if v[0] >= self.minSup and v[1] <= self.maxPer}
        plist = [key for key, value in sorted(candidate.items(), key=lambda x: (x[1][0], x[0]), reverse=True)]
        return plist

    def convert(self, value):
        """
        To convert the given user specified value
        :param value: user specified value
        :return: converted value
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self.Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self.Database) * value)
            else:
                value = int(value)
        return value

    def getSupportAndPeriod(self, timeStamps):
        """calculates the support and periodicity with list of timestamps

            :param timeStamps: timestamps of a pattern
            :type timeStamps: list
        """
        timeStamps.sort()
        cur = 0
        per = 0
        sup = 0
        for j in range(len(timeStamps)):
            per = max(per, timeStamps[j] - cur)
            if per > self.maxPer:
                return [0, 0]
            cur = timeStamps[j]
            sup += 1
        per = max(per, self.lno - cur)
        return [sup, per]

    def save(self, prefix, suffix, tidSetX):
        """Saves the patterns that satisfy the periodic frequent property.

            :param prefix: the prefix of a pattern
            :type prefix: list or None
            :param suffix: the suffix of a patterns
            :type suffix: list
            :param tidSetX: the timestamp of a patterns
            :type tidSetX: list


        """
        if prefix == None:
            prefix = suffix
        else:
            prefix = prefix + suffix
        val = self.getSupportAndPeriod(tidSetX)
        if val[0] >= self.minSup and val[1] <= self.maxPer:
            self.finalPatterns[tuple(prefix)] = val

    def Generation(self, prefix, itemSets, tidSets):
        if len(itemSets) == 1:
            i = itemSets[0]
            tidI = tidSets[0]
            self.save(prefix, [i], tidI)
            return
        for i in range(len(itemSets)):
            itemX = itemSets[i]
            if itemX == None:
                continue
            tidSetX = tidSets[i]
            classItemSets = []
            classTidSets = []
            itemSetX = [itemX]
            neighboursItemsI = self.getNeighbourItems(itemSets[i])
            for j in range(i + 1, len(itemSets)):
                neighboursItemsJ = self.getNeighbourItems(itemSets[i])
                if not itemSets[j] in neighboursItemsI:
                    continue
                itemJ = itemSets[j]
                tidSetJ = tidSets[j]
                y = list(set(tidSetX).intersection(tidSetJ))
                if len(y) >= self.minSup:
                    ne = list(set(neighboursItemsI).intersection(neighboursItemsJ))
                    x = []
                    x = x + [itemX]
                    x = x + [itemJ]
                    self.NeighboursMap[tuple(x)] = ne
                    classItemSets.append(itemJ)
                    classTidSets.append(y)
            newPrefix = list(set(itemSetX)) + prefix
            self.Generation(newPrefix, classItemSets, classTidSets)
            self.save(prefix, list(set(itemSetX)), tidSetX)

    def getNeighbourItems(self, keySet):
        """
            A function to get Neighbours of a item
            :param keySet:itemSet
            :type keySet:str or tuple
            :return: set of common neighbours 
            :rtype:set
        """
        itemNeighbours = self.NeighboursMap.keys()
        if isinstance(keySet, str):
            if self.NeighboursMap.get(keySet) is None:
                return []
            itemNeighbours = list(set(itemNeighbours).intersection(set(self.NeighboursMap.get(keySet))))
        if isinstance(keySet, tuple):
            keySet = list(keySet)
            for j in range(0, len(keySet)):
                i = keySet[j]
                itemNeighbours = list(set(itemNeighbours).intersection(set(self.NeighboursMap.get(i))))
        return itemNeighbours

    def mapNeighbours(self):
        """
            A function to map items to their Neighbours
        """
        self.NeighboursMap = []
        if isinstance(self.iFile, pd.DataFrame):
            data = []
            if self.iFile.empty:
                print("its empty..")
            i = self.iFile.columns.values.tolist()
            if 'Neighbours' in i:
                data = self.iFile['Neighbours'].tolist()
            for i in data:
                self.NeighboursMap[i[0]] = i[1:]
        if isinstance(self.iFile, str):
            if validators.url(self.iFile):
                data = urlopen(self.iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self.sep)]
                    temp = [x for x in temp if x]
                    self.NeighboursMap[temp[0]] = temp[1:]
            else:
                try:
                    with open(self.iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self.sep)]
                            temp = [x for x in temp if x]
                            self.NeighboursMap[temp[0]] = temp[1:]
                except IOError:
                    print("File Not Found")
                    quit()

    def startMine(self):
        """Frequent pattern mining process will start from here"""

        # global items_sets, endTime, startTime
        self.startTime = time.time()
        if self.iFile is None:
            raise Exception("Please enter the file path or file name:")
        iFileName = self.iFile
        self.creatingItemSets(iFileName)
        self.minSup = self.convert(self.minSup)
        self.mapNeighbours()
        self.finalPatterns = {}
        plist = self.frequentOneItem()
        for i in range(len(plist)):
            itemX = plist[i]
            tidSetX = self.tidList[itemX]
            itemSetX = [itemX]
            itemSets = []
            tidSets = []
            neighboursItems = self.getNeighbourItems(plist[i])
            for j in range(i + 1, len(plist)):
                if not plist[j] in neighboursItems:
                    continue
                itemJ = plist[j]
                tidSetJ = self.tidList[itemJ]
                y1 = list(set(tidSetX).intersection(tidSetJ))
                if len(y1) >= self.minSup:
                    itemSets.append(itemJ)
                    tidSets.append(y1)
            self.Generation(itemSetX, itemSets, tidSets)
            self.save(None, itemSetX, tidSetX)
        self.endTime = time.time()
        process = psutil.Process(os.getpid())
        self.memoryUSS = float()
        self.memoryRSS = float()
        self.memoryUSS = process.memory_full_info().uss
        self.memoryRSS = process.memory_info().rss
        print("Spatial Periodic Frequent patterns were generated successfully using SpatialEclat algorithm")

    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function
        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self.memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function
        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self.memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process
        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self.endTime - self.startTime

    def getPatternsAsDataFrames(self):
        """Storing final frequent patterns in a dataframe
        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self.finalPatterns.items():
            data.append([a, b[0], b[1]])
            dataFrame = pd.DataFrame(data, columns=['Patterns', 'Support', 'Period'])
        return dataFrame

    def savePatterns(self, outFile):
        """Complete set of frequent patterns will be loaded in to a output file
        :param outFile: name of the output file
        :type outFile: file
        """
        self.oFile = outFile
        writer = open(self.oFile, 'w+')
        for x, y in self.finalPatterns.items():
            pat = ""
            for i in x:
                pat += str(i) + " "
            patternsAndSupport = pat + ": " + str(y[0]) + ": " + str(y[1])
            writer.write("%s \n" % patternsAndSupport)

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process
        :return: returning frequent patterns
        :rtype: dict
        """
        return self.finalPatterns


if __name__ == "__main__":
    ap = str()
    if len(sys.argv) == 6 or len(sys.argv) == 7:
        if len(sys.argv) == 7:
            ap = PFS_ECLAT(sys.argv[1], sys.argv[3], sys.argv[4], sys.argv[5], sys.argv[6])
        if len(sys.argv) == 6:
            ap = PFS_ECLAT(sys.argv[1], sys.argv[3], sys.argv[4], sys.argv[5])
        ap.startMine()
        spatialFrequentPatterns = ap.getPatterns()
        print("Total number of Spatial Frequent Patterns:", len(spatialFrequentPatterns))
        ap.savePatterns(sys.argv[2])
        memUSS = ap.getMemoryUSS()
        print("Total Memory in USS:", memUSS)
        memRSS = ap.getMemoryRSS()
        print("Total Memory in RSS", memRSS)
        run = ap.getRuntime()
        print("Total ExecutionTime in seconds:", run)
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")



#  Copyright (C)  2021 Rage Uday Kiran
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation, either version 3 of the License, or
#      (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
import time
import validators
from urllib.request import urlopen
import csv
import pandas as pd
from collections import defaultdict
from itertools import combinations as c
import os
import os.path
import psutil
import sys
import validators
from urllib.request import urlopen


class utilityPatterns(ABC):
    """ This abstract base class defines the variables and methods that every topk spatial high utility pattern mining algorithm must
        employ in PAMI


    Attributes :
    ----------
        iFile : str
            Input file name or path of the input file
        k: integer
            The user can specify k (top-k)
        sep : str
            This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
            However, the users can override their default separator
        startTime:float
            To record the start time of the algorithm
        endTime:float
            To record the completion time of the algorithm
        finalPatterns: dict
            Storing the complete set of patterns in a dictionary variable
        oFile : str
            Name of the output file to store complete set of frequent patterns
        memoryUSS : float
            To store the total amount of USS memory consumed by the program
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program

    Methods :
    -------
        startMine()
            Calling this function will start the actual mining process
        getPatterns()
            This function will output all interesting patterns discovered by an algorithm
        savePatterns(oFile)
            This function will store the discovered patterns in an output file specified by the user
        getPatternsAsDataFrame()
            The function outputs the patterns generated by an algorithm as a data frame
        getMemoryUSS()
            This function outputs the total amount of USS memory consumed by a mining algorithm
        getMemoryRSS()
            This function outputs the total amount of RSS memory consumed by a mining algorithm
        getRuntime()
            This function outputs the total runtime of a mining algorithm
    """

    def __init__(self, iFile, nFile, k, sep="\t"):
        """
        :param iFile: Input file name or path of the input file
        :type iFile: str
        :param nFile: Input file name or path of the neighbourhood file
        :type nFile: str
        :param k: The user can specify k in count
        :type k: int
        :param sep: separator used to distinguish items from each other. The default separator is tab space. However, users can override the default separator
        :type sep: str
        """

        self.iFile = iFile
        self.sep = sep
        self.nFile = nFile
        self.k = k

    @abstractmethod
    def iFile(self):
        """Variable to store the input file path/file name"""
        pass

    @abstractmethod
    def nFile(self):
        """Variable to store the neighbourhood file path/file name"""
        pass

    @abstractmethod
    def startTime(self):
        """Variable to store the start time of the mining process"""
        pass

    @abstractmethod
    def endTime(self):
        """Variable to store the end time of the complete program"""
        pass

    @abstractmethod
    def memoryUSS(self):
        """Variable to store USS memory consumed by the program"""
        pass

    @abstractmethod
    def memoryRSS(self):
        """Variable to store RSS memory consumed by the program"""
        pass

    @abstractmethod
    def finalPatterns(self):
        """Variable to store the complete set of patterns in a dictionary"""
        pass

    @abstractmethod
    def oFile(self):
        """Variable to store the name of the output file to store the complete set of frequent patterns"""
        pass

    @abstractmethod
    def startMine(self):
        """Code for the mining process will start from this function"""
        pass

    @abstractmethod
    def getPatterns(self):
        """Complete set of patterns generated will be retrieved from this function"""
        pass

    @abstractmethod
    def savePatterns(self, oFile):
        """Complete set of patterns will be saved in to an output file from this function
        :param oFile: Name of the output file
        :type oFile: file
        """
        pass

    @abstractmethod
    def getPatternsAsDataFrame(self):
        """Complete set of generated patterns will be loaded in to data frame from this function"""
        pass

    @abstractmethod
    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the program will be retrieved from this function"""
        pass

    @abstractmethod
    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the program will be retrieved from this function"""
        pass


    @abstractmethod
    def getRuntime(self):
        """Total amount of runtime taken by the program will be retrieved from this function"""
        pass



from abc import ABC, abstractmethod
import time
import math
import csv
import pandas as pd
from collections import defaultdict
from itertools import combinations as c
import os
import os.path
import psutil
import sys
import validators
from urllib.request import urlopen


class partialPeriodicPatterns(ABC):
    """ This abstract base class defines the variables and methods that every frequent pattern mining algorithm must
    employ in PAMI

        ...

        Attributes
        ----------
        iFile : str
            Input file name or path of the input file
        periodicSupport: float or int or str
            The user can specify periodicSupport either in count or proportion of database size.
            If the program detects the data type of periodicSupport is integer, then it treats periodicSupport is expressed in count.
            Otherwise, it will be treated as float.
            Example: periodicSupport=10 will be treated as integer, while periodicSupport=10.0 will be treated as float
        period: float or int or str
            The user can specify period either in count or proportion of database size.
            If the program detects the data type of period is integer, then it treats period is expressed in count.
            Otherwise, it will be treated as float.
            Example: period=10 will be treated as integer, while period=10.0 will be treated as float
        sep : str
            This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
            However, the users can override their default separator.
        startTime:float
            To record the start time of the algorithm
        endTime:float
            To record the completion time of the algorithm
        finalPatterns: dict
            Storing the complete set of patterns in a dictionary variable
        oFile : str
            Name of the output file or path of the output file
        memoryUSS : float
            To store the total amount of USS memory consumed by the program
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program

        Methods
        -------
        startMine()
            Mining process will start from here
        getFrequentPatterns()
            Complete set of patterns will be retrieved with this function
        savePatterns(oFile)
            Complete set of frequent patterns will be loaded in to a output file
        getPatternsAsDataFrame()
            Complete set of frequent patterns will be loaded in to data frame
        getMemoryUSS()
            Total amount of USS memory consumed by the program will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the program will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the program will be retrieved from this function
    """

    def __init__(self, iFile, periodicSupport, period, sep = '\t'):
        """
        :param iFile: Input file name or path of the input file
        :type iFile: str
        :param minSup: UserSpecified minimum support value. It has to be given in terms of count of total number of
        transactions in the input database/file
        :type minSup: float
        """

        self.iFile = iFile
        self.periodicSupport = periodicSupport
        self.period = period
        self.sep = sep

    @abstractmethod
    def iFile(self):
        """Variable to store the input file path/file name"""

        pass

    @abstractmethod
    def periodicSupport(self):
        """Variable to store the user-specified minimum support value"""

        pass
    def period(self):
        """Variable to store the user specified maximum periodicity value"""

        pass

    @abstractmethod
    def sep(self):
        """Variable to store the input file path/file name"""

        pass

    @abstractmethod
    def startTime(self):
        """Variable to store the start time of the mining process"""

        pass

    @abstractmethod
    def endTime(self):
        """Variable to store the end time of the complete program"""

        pass

    @abstractmethod
    def memoryUSS(self):
        """Variable to store the end time of the complete program"""

        pass

    @abstractmethod
    def memoryRSS(self):
        """Variable to store the end time of the complete program"""

        pass

    @abstractmethod
    def finalPatterns(self):
        """Variable to store the complete set of patterns in a dictionary"""

        pass

    @abstractmethod
    def oFile(self):
        """Variable to store the name of the output file to store the complete set of frequent patterns"""

        pass

    @abstractmethod
    def startMine(self):
        """Code for the mining process will start from this function"""

        pass

    @abstractmethod
    def getPatterns(self):
        """Complete set of frequent patterns generated will be retrieved from this function"""

        pass

    @abstractmethod
    def savePatterns(self, oFile):
        """Complete set of frequent patterns will be saved in to an output file from this function

        :param oFile: Name of the output file
        :type oFile: file
        """

        pass

    @abstractmethod
    def getPatternsAsDataFrame(self):
        """Complete set of frequent patterns will be loaded in to data frame from this function"""

        pass

    @abstractmethod
    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the program will be retrieved from this function"""

        pass

    @abstractmethod
    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the program will be retrieved from this function"""
        pass

    @abstractmethod
    def getRuntime(self):
        """Total amount of runtime taken by the program will be retrieved from this function"""

        pass
